"""
Output file format ("*.mat"):

First line has 3 integers::

  -- I: number of rows in matrix (= number of words in vocab)
  -- J: number of columns in the matrix ( = number of deprel types, len of
  deprel codebook)
  -- T: total number of nonzero entries in the I x J matrix ( = total
     number of deprel tokens)

From the Cluto manual:

non-zero entries from a are specified as space-separated pairs.
Each pair contains the col number followed by the value for
that particular column (ie. feature).  The col numbers are
assumed to be integers and their corresponding values are
aasusmed to be floats. An example:

7 8 21

2 1.1   5 -0.5      8 0.2
1 1.4   2  0.4      4 -0.4
3 1.8   6  2.0      8 3.0
1 1.0
2 5.5   4 3.0       7 8.0
3 1.0   5 -1.0      6 2.0
2 3.5   4 -1.0      5 4.0   7 2.0   8 8.0

End quote

Example call.  Make cluto file for the 500 most frequent nouns::

   >>> write_cluto_file(dependency_train_file,word_deprel_ct_dict, pos='NN',top_n=500)

Shd also independently/optionally make an rlabel (= row label) file:

  >>> make_row_label_file(dependency_train_file,pos='NN',top_n=500)

For the next commandline to work, the two output files, dependency_DB_500_NN.mat,
and dependency_DB_500_NN.mat.rlabel shd be copied to $cluto/Matrices/.
As written the next command is administered from the $cluto directory (which has binaries for various machine architectures stored
in various subdirectories, here Darwin-i386 binaries ar used)::

Darwin-i386/vcluster -rlabelfile=Matrices/dependency_DB_500_NN.mat.rlabel -clustfile=Matrices/dep_500_nn.cluster Matrices/dependency_DB_500_NN.mat 20

20 clusters are made using the input data file, "Matrices/dependency_DB_500_NN.mat".

"""
import os.path
import dep_file_utils

clu_ext = '.mat'

def write_cluto_file(filename, word_deprel_ct_dict, \
                     pos='', top_n=0, lexemes=None):
    """
    Write out a file in cluto spare matrix format and associated
    row-key file, based on the dependency model in C{word_deprel_ct_dict},
    which has been generated from the raw counts file C{filename}.

    If the lexemes to save are not supplied as the value of the parameter C{lexemes}
    we build C{lex_list}, a (possibly sorted) item list of all
    lexemes paired with their counts.  We iterate thorugh it twice:
    first, to extract lex & feat (row & column) stats required by cluto and (possibly)
    an indexing scheme to renumber features (when we are clustering
    only a subset of the data); second, to actually
    write out a file in cluto sparse-matrix format.

    When we extract a subset of the data by frequency, we need to
    open the original counts file, parameter C{filename} to get the
    raw counts.
    """
    global lexemes1, ctr, total_entry_ct, feat_encodings, lex_list
    
    (base, ext) = os.path.splitext(filename)
    (pos_str, top_n_str) = dep_file_utils.get_file_suffixes(pos,  top_n)
    ofh = open(base + top_n_str + pos_str + clu_ext ,'w')

    if top_n and not lexemes:
        lex_list = dep_file_utils.get_vocab(filename,pos,top_n)
    elif not lexemes:
        lex_list = dep_file_utils.word_ct_dict.items()
        if lex_list == [ ]:
            print 'Using deprel ct dict for lex_list'
            lex_list = word_deprel_ct_dict.items()
    if not lexemes:
        lexemes = []
        build_lexemes = True
    else:
        lex_list = lexemes
        build_lexemes = False
        
    feat_encodings = {}
    feat_ctr = dep_file_utils.Ctr()
    total_entry_ct = 0
    ctr = 0
    for pair in lex_list:
        if build_lexemes:
            ((lex, lpos), ct) = pair
        else:
            (lex, lpos) = pair
        if pos and lpos <> pos:
            continue
        ctr += 1
        if top_n and ctr > top_n:
            break
        if build_lexemes:
            lexemes.append((lex,lpos))
        this_dict = word_deprel_ct_dict[(lex,lpos)]
        for deprel in this_dict:
            total_entry_ct += 1
            encode_feat(deprel,feat_encodings,feat_ctr)
    # Save to this global var for debugging
    lexemes1 = lexemes
    print 'Writing Cluto training file'
    print >> ofh, ctr, feat_ctr.get_val(), total_entry_ct
    ctr = 0
    for (word,lpos) in lexemes:
        this_dict = word_deprel_ct_dict[(word,lpos)]
        keylist = this_dict.keys()
        keylist.sort()
        for deprel in keylist:
            print >> ofh, '%d %.8f' % (feat_encodings[deprel],\
                                        this_dict[deprel]),
        print >> ofh
    ofh.close()
    print 'Cluto matrix file written'
    key_file = base + top_n_str + pos_str + '_key' + '.clu'
    write_key_file(key_file, lexemes, pos, top_n)
    if pos or top_n:
        feat_encoding_file = base + top_n_str + pos_str + '_feats' + '.clu'
        write_feat_encodings(feat_encoding_file, feat_encodings)

def encode_feat(feat, feat_encodings,feat_ctr):
    if feat in feat_encodings:
        return feat_encodings[feat]
    else:
        next_val = feat_ctr.get_next()
        feat_encodings[feat] = next_val
        return next_val


def write_feat_encodings(feat_encoding_file, feat_encodings):
    print 'Writing Cluto feat_encodings to %s' % (feat_encoding_file,)
    try:
        fe_fh = open(feat_encoding_file, 'w')
        for feat in feat_encodings:
            print >> fe_fh, feat, feat_encodings[feat]
    finally:
        fe_fh.close()
    print 'Cluto feat encodings written'

def write_key_file(key_file, lexemes, pos=None, top_n=None):
    print 'Writing Cluto matrix keys to %s' % (key_file,)
    ctr = 0
    try:
        kfh = open(key_file, 'w')
        for (index,(lex,lpos)) in enumerate(lexemes):
            if pos and lpos <> pos:
                continue
            ctr += 1
            if top_n and ctr > top_n:
                break
            # Add 1 so we print actual row number in Cluto matrix file (line_no -1)
            print >> kfh, index+1, lex, lpos
    finally:
        kfh.close()
    print 'Cluto matrix keys written'

def make_row_label_file (filename, pos=None, top_n=None, lexemes=None):
    """
    New version
    """
    
    (base, ext) = os.path.splitext(filename)
    (pos_str, top_n_str) = dep_file_utils.get_file_suffixes(pos,  top_n)
    key_file = base + top_n_str + pos_str + clu_ext + '.rlabel'
    if lexemes is None:
        lexemes = dep_file_utils.get_vocab(filename, pos, top_n)
        lexemes = [x[1] for x in lexemes]
    write_row_label_file(key_file, lexemes, pos=pos, top_n=top_n)
    
def write_row_label_file(key_file, lexemes, pos=None, top_n=None):
    """
    If pos is fixed we dont need to incloude pos in the row
    label file so in the interest of short labels, we dont.
    """
    print 'Writing Cluto format row label file to %s' % (key_file,)
    ctr = 0
    try:
        kfh = open(key_file, 'w')
        for (lex,lpos) in lexemes:
            if pos and lpos <> pos:
                continue
            ctr += 1
            if top_n and ctr > top_n:
                break
            if pos:
                print >> kfh, lex
            else:
                print >> kfh, '%s_%s' % (lex, lpos)
    finally:
        kfh.close()
    print 'Cluto row labels file written'

##############  Moved to cluster_tree.py ############################################
# def get_clusters (key_file, cluster_file, num_clusters, display=True):
#     clusters = {}
#     for k in range(-1,num_clusters):
#         clusters[k] = []
#     kfh = open(key_file,'r')
#     cfh = open(cluster_file,'r')
#     for line in cfh:
#         clust = int(line.strip())
#         lexeme = kfh.readline().strip()
#         clusters[clust].append(lexeme)
#     kfh.close()
#     cfh.close()
#     if display:
#         print_clusters (clusters)
#     return clusters
# 
# ##############  Moved to cluster_tree.py ############################################
# ## get_tree('cluto/dependency_DB_500_NN_key.clu',   'cluto/dep_500_nn.cluster', 50, 'cluto/dependency_DB_500_NN.mat.cltree.50')
# def get_tree (key_file, cluster_file, num_clusters, tree_file):
#     clusters = (key_file, cluster_file, num_clusters)
#     tfh = open(tree_file,'r')
#     parents = []
#     children = {}
#     ctr = 0
#     for line in tfh:
#         print line
#         parent = int(line.strip().split()[0])
#         print parent
#         parents.append(parent)
#         if parent in childredn:
#             children[parent].append(ctr)
#         else:
#             children[parent] = [ctr]
#         ctr += 1
#         
# def print_clusters (clusters):
#     for k in clusters:
#         print 'Cluster %d:' % k
#         for lex in clusters[k]:
#             print '    %s' % lex    
# 
        
