"""
Notes on the sparse matrix implementation. [See http://www.scipy.org/SciPyPackages/Sparse]

>>> import scipy.sparse
>>> import numpy as np

We build a 1x2 sparse matrix  [ a vector ]
>>> row = np.array([0,0])
>>> col = np.array([0,1])
>>> data = np.array([2,7])
>>> X = scipy.sparse.csc_matrix((data,(row,col)), shape=(1,2))

csr sparse matrix used here. One of 7 types.  Inefficient for
updating, though this is allowed [X[0,0] = <val>]
>>> X
<1x2 sparse matrix of type '<type 'numpy.int64'>'
	with 2 stored elements in Compressed Sparse Column format>
>>> X.todense()
matrix([[2, 7]])

Now a Y suitable for Matrix multiplication. (2x1)
>>> Y = scipy.sparse.csc_matrix((data,(col,row)), shape=(2,1))
>>> Y.todense()
matrix([[2],
        [7]])
>>> X * Y
<1x1 sparse matrix of type '<type 'numpy.int64'>'
	with 1 stored elements in Compressed Sparse Column format>
>>> Z = X * Y
>>> X.todense()
matrix([[2, 7]])
>>> Z.todense()
matrix([[53]])

So if we have a bunch of word vectors represented as sparse 0xn
vectors (where n is the number of word features), then dot product is defined
as follows.

def dot_prod (vec1, vec2):
   return vec1.transpose().dot(vec2).todense()[0,0]

arithmetic ops including * are suppsoed tobe fast for CSR format spare matrices,
but C{todense} is supposed to be expensive, so csr may not be the best
choice of scipy's 7 possible sparse formats,
but it may not matter if we only ever do todense() for 1x1 matrices.

"""
import math, os.path, time
from collections import Counter
old_dir = os.getcwd()
home = os.getenv('HOME')
#home = '/Users/gawron/'
xdependency_db_dir = os.path.join(home, 'Ranking/RerankingPackage/dependency_DB')
os.chdir(xdependency_db_dir)
import sim_measures
os.chdir(old_dir)
#from sim_measures import counts_to_mutual_info, counts_to_mutual_info_nonneg, counts_to_pointwise_mutual_info, counts_to_t_test, counts_to_t_test_nonneg, IRad, conditional_probs, lin, lin_no_neg, lin_no_neg2, lin_no_neg3, lin_no_neg_principled, dot_prod, dot_prod_no_neg, jaccard,counts_to_undiscounted_pointwise_mutual_info, counts_to_pointwise_mutual_info_nonneg, sqrt_freq, counts_to_info, dice_dagger, dice_dagger_no_neg,gr_counts_to_pointwise_mutual_info
import dep_file_utils, dep_rel_info
## A global var shared among modules. Probably a better way to do this.
from dep_file_utils import inverse_deprel_char as inv_char
import nltk
import nltk.corpus.reader.wordnet as wn
from bnc_wn_interface import bnc_pos_to_wn_pos

######################## Global Vars #############################

## lemma_dict
# lemma will be a dotted pair (wn- and fn- style) of a citation form and a wn pos ('v' or 'n')
# lemma_dict has lemma keys. and sets of (form,bnc_pos) pairs as values.
# Despite having a degree in linguistics, I have been inacurately calling such pairs lexemes. For example:
# lemma_dict['be.v'] = set([('be','VB'),('being','VBG'),('been','VBN'),('is','VBZ'),('was','VBD'),('am','VBP'),('are','VBP')])
## word_deprel_ct_dict: Every lexeme, an  orthography(str),pos(str) pair, is associated with
## a dictionary of deprel counts.  So this one is a dict of dicts.

## deprel_ct_dict: Every dep_rel [an int] associated with a count

## word_ct_dict: total number of deprels for each word
## Every lexeme [an  orthography(str),pos(str) pair] associated with a count

(word_deprel_ct_dict, deprel_ct_dict, word_ct_dict, lemma_dict, lexeme_deprel_ct_dict, lexeme_ct_dict) = ({ }, { }, { }, { },{ },{ })

# Safe versions are to be used a backup copies, debugging.
(safe_word_deprel_ct_dict, safe_deprel_ct_dict, safe_word_ct_dict, safe_lemma_dict,safe_lexeme_deprel_ct_dict, safe_lexeme_ct_dict ) = ({ }, { }, { }, { }, { }, { })


(nearest_neighbor_dict, rel_nearest_neighbor_dict) = ({ }, { })

## nearest_neighbor_dict: keys are lexemes. vals are lexeme, dotprod value pairs.

## rel_nearest_neighbor_dict: keys are pos. Vals are dictionaries. rel_nearest_neighbor['NN']
## a nearest neighbor dict relating lexemes of ANY pos to the nearest NN neighbor.
## rel_nearest_neighbor['NN'][(wishes,'VBZ') is the nearest noun neighbor of
## "wishes'

## a list of dependency_triples used for decoding integer vals in dependency_db_keys
## file. decodebook[index] = (form, pos, gf) triple
#decodebook = [ ]

## mutual_nearest_neighbor list
mutual_nearest_neighbors = [ ]

## words that show up with no dependencies
misfit_list = [ ]

######################## End Global Vars ###########################

def make_pruned_model(dependency_train_file,dependency_keys_file,\
                      decodebook, feat_threshhold=4, wd_threshhold=8,\
                      target_feat_num = None, prune_neg=False):
    """
    Stanford BNC::

     len(word_deprel_ctr_dict) = 913812
     len(deprel_ct_dict) = 6296780

     pruned.  560784  thresh=2
     num_floats.get_val() = 56,179,752
     numb bytes = 56,179,752  * 8 = 449,438,016

     =======================================================
     feat_thresh = 2, wd_threshold=6?
     =======================================================
     3670510 pruned feats
     241190 pruned words
     3895865 pruned cells
     312427 lexical items were saved, 360195 pruned squares saved: False

     =======================================================
     feat_thresh = 4,wd_threshhold=8 (8 million cells = 14%)
     =======================================================
     
     4,945,449 pruned feats
     6,654,823 pruned cells
     1,454,246 cells pruned by word pruning
     262730 lexical items were saved; 651082 were pruned;

    Starting lex size: 915305 => 262730

    It makes sense to prune more words but not necessarily
    prune their feats.  That way we save on memory but dont lose
    resolution in the model::
      wd_threshhold: wordcount at which to prune a word, can be high
         a parameeter here and done in output_dependency_file.
         Passed on to it.
      feat_wd_threshold:  word ct for word w at which we prune all
         feats of the form (w,f). Loses model resolution. Keep low.
      feat_word_threshhold:  refers to ct of teh number of distinct
         words a feat occurs with.
      
    """
    global pruned_feats,non_word_ctr, feat_word_freq #,pruned_words
    print 'reading %s' % (dependency_train_file,)
    # set global feat_word_freq,  For legacy reasons not returning it.
    (word_deprel_ct_dict, word_ct_dict, deprel_ct_dict, \
     lemma_dict, total_dep_ct) = \
    read_dependency_file(dependency_train_file, count_file=True,
                         suffix='', ext = '')
    print 'getting decode book %s' % (dependency_keys_file,)
    print 'Finding low value feats'
    num_feats_left = len(feat_word_freq)
    if target_feat_num is None:
       target_feat_num = num_feats_left - 1
    while num_feats_left > target_feat_num:
        (pruned_feats, non_word_ctr, num_pruned_feats,num_feats_left) = \
             find_prune_feats(decodebook,word_ct_dict,feat_word_freq,
                              feat_threshhold, wd_threshhold)
        feat_threshhold += 1
    pruned_cells = 0
    for w in word_deprel_ct_dict.keys():
        fd = word_deprel_ct_dict[w]
        for f in fd.keys():
            if f in pruned_feats:
                pruned_cells += 1
                del fd[f]
            elif prune_neg and fd[f] < 0:
                pruned_cells += 1
                del fd[f]
    print '%d pruned cells' % pruned_cells
    if prune_neg:
        ext = '_nonneg'
    else:
        ext = ''
    output_dependency_dict(word_deprel_ct_dict, dependency_train_file, \
                           suffix='',  ext=ext, prune=True,count_file=True,\
                           threshhold=wd_threshhold,word_ct_dict=word_ct_dict)
    return pruned_feats
            
        
def find_prune_feats(decodebook,word_ct_dict,feat_word_freq,
                     feat_threshhold, wd_threshhold):

    (pruned_feats,all_feats,non_word_ctr) = (set(),set(),Counter())
    # omit first bogus feat
    for dep_rel_code in range(1, len(decodebook)):
        (wd,dep_rel,pos) = decodebook[dep_rel_code]
        all_feats.add(dep_rel_code)
        try:
            word_ct = word_ct_dict[(wd,pos)]
        except KeyError:
            # feats corresponding to things that have been
            # declared not to be words.  1495 of these in base
            # stanford_bnc model.
            pruned_feats.add(dep_rel_code)
            non_word_ctr[(wd,pos)] += 1
            continue
        if word_ct < wd_threshhold:
            pruned_feats.add(dep_rel_code)
            #pruned_words.add((wd,pos))
        elif feat_word_freq[dep_rel_code] < feat_threshhold:
            pruned_feats.add(dep_rel_code)
    num_pruned_feats = len(pruned_feats)
    num_feats_left = len(all_feats) - num_pruned_feats
    print '%d feat threshhold:' % feat_threshhold
    print '    %d pruned feats' % num_pruned_feats
    print '    %d feats left' % num_feats_left 
    return  (pruned_feats, non_word_ctr, num_pruned_feats,num_feats_left)



def convert_to_mi (dependency_train_file, \
                   mi_func = 'counts_to_mutual_info', normalize='l2',\
                   dependency_keys_file= None,\
                   suffix='', ext='',lemmatizer=False, save_cts=False,cts_only=False,
                   new_wt_file='',write_dir='', store_sqs=False):
    """
    Convert a dependency DB containing raw counts to one containing weighted association values.

    When C{normalize} = 'l2', normalize word vector lengths to
    be 1, to facilitate using cosine as sim measure [now just the dot product of 2 vectors].
    When C{normalize} = 'l1', normalize word dependency vector sums to
    be 1, dividing by the sum of the features. For each feature f of word w, the value  may now be thought of
    as a conditional probability, pr(f|w).

    C{mi_func} is a STRING to be converted to a function obj  through the sim_measures module, so it
    must be the name of a function defined in that module.
    """
    global word_deprel_ct_dict, word_ct_dict, deprel_ct_dict, lemma_dict, \
           safe_word_deprel_ct_dict, safe_word_ct_dict, safe_deprel_ct_dict, safe_lemma_dict,\
           lexeme_deprel_ct_dict,lexeme_ct_dict, safe_count_dicts, dependent_ct_dict, \
           head_ct_dict

    count_dicts = [word_deprel_ct_dict, word_ct_dict, deprel_ct_dict, lemma_dict]
    safe_count_dicts = [safe_word_deprel_ct_dict, safe_word_ct_dict, safe_deprel_ct_dict, safe_lemma_dict]
    zipped_dicts = zip(count_dicts,safe_count_dicts)

    (base,f_ext) = os.path.splitext(dependency_train_file)
    if dependency_keys_file is None:
        dependency_keys_file = base + '.keys'

    # Note this avoids the soon to be deprecated from sim_measures import *
    # But note this also means other modules calling this fn can use string names!
    mi_func = getattr(sim_measures, mi_func)
    sim_measures.test_norm_constraint(mi_func,normalize)

    print 'Reading dependency file'
    # Note WE DONT pass in C{suffix} and C{ext}. Those are arguments of read_dependency
    # file ONLY when reading in a dep file which has MI numbers (rather than counts).
    # Here we read in the original count file to MAKE an mi file, and so we
    # use the original unsuffixed name.
    (word_deprel_ct_dict, word_ct_dict, \
     deprel_ct_dict, lemma_dict, total_deprels) = \
         read_dependency_file(dependency_train_file,lemmatizer=lemmatizer)
    print '%s read' % (dependency_train_file,)
    if lemmatizer:
        (lexeme_deprel_ct_dict,lexeme_ct_dict) = \
                       merge_word_form_counts(word_deprel_ct_dict, word_ct_dict, lemma_dict)
        ## Need a real lexeme count file for various kinds of stat manipulations
        ## This only need be done once.
        print 'Writing out lexeme counts file'
        output_dependency_dict(lexeme_deprel_ct_dict, dependency_train_file, \
                               suffix='',  ext='', prefix='lemma_', count_file=True)
    if cts_only:
        return
    if save_cts:
        for (d,safe_d) in zipped_dicts:
            for (k,v) in d.iteritems():
                if type(v) == type(d):
                    this_dict = {}
                    this_dict.update(v)
                    safe_d[k] = this_dict
                else:
                    safe_d[k] = v
    if lemmatizer:
        print 'Converting lexeme counts to association weights'
        convert_word_deprel_ct_dict_to_mi (lexeme_deprel_ct_dict, lexeme_ct_dict, \
                                           deprel_ct_dict, total_deprels, \
                                           mi_func, normalize)
        output_dependency_dict(lexeme_deprel_ct_dict, dependency_train_file, \
                               suffix=suffix,  ext=ext, prefix='lemma_')

    if store_sqs:
        sqs_dict = {}
    else:
        sqs_dict = None
    (rel_dict, decodebook) = dep_rel_info.get_deprel_info(dependency_keys_file)

    print 'Converting word counts to association weights'
    convert_word_deprel_ct_dict_to_mi (word_deprel_ct_dict, word_ct_dict, \
                                       deprel_ct_dict, total_deprels, \
                                       mi_func, normalize, store_sqs=store_sqs,\
                                       sqs_dict=sqs_dict,decodebook=decodebook)
    output_dependency_dict(word_deprel_ct_dict, dependency_train_file, \
                           suffix=suffix,  ext=ext,\
                           new_wt_file=new_wt_file, write_dir=write_dir,\
                           store_sqs=store_sqs,sqs_dict=sqs_dict)


def read_dependency_file(dependency_file, count_file=True, suffix='',
                         ext = '', lemmatizer=False, prefix=''):

    """
    Open dependendency DB file and populate 3 count dictionaries, word-to-word,
    dep cts,  word cts, and dep cts.  Also compute total number of deprels [= total
    num of words in corpus - 1] and return it.

    Note: The combination lemmatizer=True and count_file = False is an error.  We use
    the lemmatizer for summing lex form counts into lexeme counts.  This makes no
    sense with non count vals like MI numbers, PMI numbers, or conditional probs.
    """
    global lemma_ctr

    if lemmatizer:
        assert count_file, "Can't use lemmatizer when reading in a non count dependency file!"
    if prefix:
        (path,filename) = os.path.split(dependency_file)
        dependency_file = os.path.join(path,prefix+filename)
    (base,final_ext) = os.path.splitext(dependency_file)
    full_dependency_file = base + suffix + ext + final_ext
    print 'Reading %s count file: %s' % (full_dependency_file, count_file)
    return read_dependency_file_helper (full_dependency_file,lemmatizer,count_file)

def read_dependency_file_helper (full_dependency_file,lemmatizer,count_file):
    global wt_file, num_floats, feat_word_freq
    fh = open(full_dependency_file,'r')
    total_deprels = dep_file_utils.Ctr()
    num_floats = dep_file_utils.Ctr()
    feat_word_freq = Counter()
    lemma_dict = {}
    lemma_ctr = 0
    if lemmatizer:
        cr = wn.WordNetCorpusReader(nltk.data.find('corpora/wordnet'))
        wnl = cr._morphy
    else:
        wnl = None
    for line in fh:
        try:
            (lexeme, dep_ct_pairs) = dep_file_utils.parse_dep_line(line, count_file=count_file)
        except TypeError:
            continue
        except ValueError:
            continue
        (wd,pos) = lexeme
        if not (pos[0] in ('N','R','J','V')):
            # Filter non major pos's: Memory help with bnc_stanford
            continue
        if dep_ct_pairs == [ ]:
            misfit_list.append(lexeme)
            token_ctr = 1
        else:
            token_ctr = 0
        this_dict = {}
        word_deprel_ct_dict[lexeme] = this_dict
        if lemmatizer:
            lemma_update(lexeme,lemma_dict, wnl)
        for (feat,count) in dep_ct_pairs:
            if count_file:
                total_deprels.inc(count)
                num_floats.inc(1)
                feat_word_freq[feat] += 1
                token_ctr += count
                dep_file_utils.update_ct_dict(deprel_ct_dict,feat,count)
            else:
                ## Here we just (mis)use this dict to represent the set of
                ## deprels.
                deprel_ct_dict[feat] = 1
            this_dict[feat] = count
        if count_file:
            dep_file_utils.update_ct_dict(word_ct_dict,lexeme,token_ctr)
        else:
            # if not counting tokens, used to represent the set of lexemes.
            # use -1 to signal this is not a count dict to other functions
            word_ct_dict[lexeme] = -1
    fh.close()
    if count_file:
        wt_file = full_dependency_file
    return (word_deprel_ct_dict, word_ct_dict, deprel_ct_dict, lemma_dict, total_deprels.get_val())


def convert_word_deprel_ct_dict_to_mi (word_deprel_ct_dict,word_ct_dict,deprel_ct_dict,\
                                       total_deprels, mi_func, normalize,\
                                       store_sqs=False,sqs_dict=None, decodebook=None):
    """
    C{word_deprel_ct_dict} is a dictionary with lexeme keys.
    Each value is dictionary representing a sparse vector of dependency counts
    for this lexeme. Convert the counts to mutual information values.

    Normalize Option = 'l2': Normalize the word dependency vector lengths to be 1,
    to faciliate using cosine as sim measure [now just the dot product of
    2 vectors].

    Normalize Option = 'l1': Normalize the word dependency vector values to add up to 1,
    to make them conditional probs.  Can only be used with mi_func = conditional_probs
    
    Normalize Option = 'hell': Like 'l1': Normalize the word dependency vector values to add up to 1,
    but do so by transforming the mi values with the sqrt fn and
    dividing by the sqrt of the sum of the mi values. Used with geometric mean.
    let  sum(w1),sum(w2) be the sum of the mi vals of vectors w1 and w2 and
    let sqrt(w1),sqrt(w2) be the elementwise sqrt of w1 and w2.
    Then
    cos(sqrt(w1),sqrt(w2)) = dotprodsqrt(w1/sum(w1),w2/sum(w2)) = dotprod(sqrt(w1/sum(w1)),sqrt(w1/sum(w1)))
"""
    global dependent_ct_dict, head_ct_dict
    
    sim_measures.test_norm_constraint(mi_func,normalize)
    ## A few last minute calculations requiring decoding, if needed
    if mi_func == sim_measures.counts_to_lin_pointwise_mutual_info:
        lin_word_deprel_prob = compute_word_deprel_probs(word_deprel_ct_dict,decodebook)
    elif mi_func == sim_measures.counts_to_hd_pointwise_mutual_info:
        (dependent_ct_dict,head_ct_dict) = get_head_dependent_cts(word_deprel_ct_dict,decodebook)
    for lexeme in word_deprel_ct_dict:
        this_dict = word_deprel_ct_dict[lexeme]
        lexeme_ct = word_ct_dict[lexeme]
        sum_squares = 0
        sum_cts = 0
        sum_mis = 0.0
        for deprel in this_dict.keys():
            class_term_ct = this_dict[deprel]
            if normalize in ['l2', '', 'hell']:
                class_non_term_ct = deprel_ct_dict[deprel] - class_term_ct
                non_class_term_ct = lexeme_ct - class_term_ct
                class_or_term_ct = class_term_ct + class_non_term_ct + non_class_term_ct
                non_class_non_term_ct = total_deprels  - class_or_term_ct
                if mi_func == sim_measures.counts_to_lin_pointwise_mutual_info:
                    this_mi = mi_func(lexeme, class_term_ct,class_non_term_ct,\
                                      non_class_term_ct, non_class_non_term_ct,\
                                      lin_word_deprel_prob[lexeme][deprel])
                elif mi_func == sim_measures.counts_to_hd_pointwise_mutual_info:
                    #old_mi = sim_measures.counts_to_pointwise_mutual_info(\
                    #                  lexeme, class_term_ct, class_non_term_ct,\
                    #                  non_class_term_ct, non_class_non_term_ct)
                    (dep_word,g_rel,d_pos) = decodebook[deprel]
                    dep_lex = (dep_word, d_pos)
                    if g_rel.startswith(dep_rel_info.inv_char):
                        lexeme_ct = dependent_ct_dict[lexeme]
                        class_ct = head_ct_dict[dep_lex]
                    else:
                        lexeme_ct = head_ct_dict[lexeme]
                        class_ct = dependent_ct_dict[dep_lex]
                    if class_ct == 0:
                        print 'class ct 0: %s %d %s %s %s' % (lexeme, deprel, dep_word,g_rel,d_pos)
                        this_mi = 0.0
                    elif lexeme_ct == 0:
                        print 'term ct 0: %s %d %s %s %s' % (lexeme, deprel, dep_word,g_rel,d_pos)
                        this_mi = 0.0
                    else:
                        class_non_term_ct = class_ct - class_term_ct
                        non_class_term_ct = lexeme_ct - class_term_ct
                        class_or_term_ct = class_term_ct + class_non_term_ct + non_class_term_ct
                        non_class_non_term_ct = total_deprels  - class_or_term_ct
                        try:
                            this_mi = mi_func(lexeme, class_term_ct,class_non_term_ct,\
                                              non_class_term_ct, non_class_non_term_ct)
                        except ZeroDivisionError, e:
                            print 'Zero Div: %s %d %d %s %s %s' % (lexeme, class_term_ct, deprel, dep_word,g_rel,d_pos)
                            raise e
                    #print 'PMI: %.7f  HD PMI: %.7f' % (old_mi, this_mi)
                else:
                    this_mi = mi_func(lexeme, class_term_ct, class_non_term_ct,\
                                      non_class_term_ct, non_class_non_term_ct)
                    
                if this_mi:
                    ## Can be 0.  Treat like other 0's: No entry.
                    this_dict[deprel] = this_mi
                    sum_mis += this_mi
                    sum_squares += this_mi**2
                else:
                    ## NB: must eliminate this, since there is still a COUNT value stored in this dictionary
                    del this_dict[deprel]
            else:
                sum_cts += class_term_ct
        # now normalize, if necessary
        if normalize == 'l2':
            norm_len = sum_squares**0.5
            for deprel in this_dict:
                val = this_dict[deprel]
                this_dict[deprel] = (val/norm_len)
        elif normalize == 'hell':
            for deprel in this_dict:
                val = this_dict[deprel]
                this_dict[deprel] = math.sqrt(val/sum_mis)
        elif normalize == 'l1':
            mi_func(this_dict,sum_cts)
        if store_sqs:
            sqs_dict[lexeme] = sum_squares


def compute_word_deprel_probs(word_deprel_ct_dict,decodebook):
    lin_word_deprel_prob = dict()

    for (lexeme,this_dict) in word_deprel_ct_dict.items():
        total_ct = 0
        word_ctr = Counter()
        deprel_ctr = Counter()
        
        for (deprel,ct) in this_dict.items():
            total_ct += ct
            (dep_word,dep_rel,d_pos) = decodebook[deprel]
            deprel_ctr[(dep_rel,d_pos)] += ct
            word_ctr[dep_word] += ct

        new_dict = dict()
        lin_word_deprel_prob[lexeme] = new_dict
        total_ct_sq = (total_ct ** 2)
        for (deprel,ct) in this_dict.items():
            (dep_word,dep_rel,d_pos) = decodebook[deprel]
            new_dict[deprel] = float(deprel_ctr[(dep_rel,d_pos)] * word_ctr[dep_word])/total_ct_sq

    return lin_word_deprel_prob

def get_head_dependent_cts(word_deprel_ct_dict,decodebook):
    (dependent_ct_dict,head_ct_dict) = (Counter(), Counter())
    for lexeme in word_deprel_ct_dict:
        this_dict = word_deprel_ct_dict[lexeme]
        for deprel in this_dict.keys():
            (dep_word,g_rel,d_pos) = decodebook[deprel]
            this_dep_ct = this_dict[deprel]
            if g_rel.startswith(dep_rel_info.inv_char):
                dependent_ct_dict[lexeme] += this_dep_ct
            else:
                head_ct_dict[lexeme] += this_dep_ct
    return (dependent_ct_dict,head_ct_dict)

def find_nearest_neighbors (dep_file, nearest_neighbor_file=None, top_n=0, pos='', \
                                preserve_pos=False, ext='', sim_func='dot_prod',\
                                deprel_dict=word_deprel_ct_dict,verbose=False, item_list=None,
                                wt_file=None, wt_func=None):
    suffix = '_' + sim_func
    sim_func = getattr(sim_measures,sim_func)
    if wt_func:
        wt_func = getattr(sim_measures,wt_func)
    return __find_nearest_neighbors__ (dep_file, nearest_neighbor_file, top_n, pos, \
                                       preserve_pos, suffix, ext, sim_func,\
                                       deprel_dict,verbose, item_list, wt_file,\
                                       wt_func)
                


def __find_nearest_neighbors__ (dep_file, nearest_neighbor_file, top_n, pos, \
                                preserve_pos, suffix, ext, sim_func,\
                                deprel_dict,verbose, item_list, wt_file,\
                                wt_func):
    global word_ct_dict
    start = time.time()
    if top_n:
        print 'top_n!', top_n
        # returns a sorted list of specified length and pos
        if item_list:
            print 'Passed in item list'
            vocab = item_list
        else:
            print 'Getting sorted vocab'
            vocab = dep_file_utils.get_vocab(dep_file,pos,top_n)
            print 'Got sorted vocab'
    else:
        if item_list:
            print 'Passed in item list: length %d' % (len(item_list),)
            vocab = item_list
        else:
            print 'Using word ct dict.'
            vocab = word_ct_dict.items()
    ctr = 0
    print 'Finding nearest neighbors'
    nn_pairs = []
    for (lexeme, ct) in vocab:
        ctr += 1
        if preserve_pos:
            pos = lexeme[1]
        before = time.time()
        nn = __find_nearest_neighbor__(lexeme, pos=pos,sim_func=sim_func,\
                                       deprel_dict=deprel_dict,wt_func=wt_func)
        after = time.time()
        if verbose:
            print lexeme, '%.5f secs' % (after - before,), '%.5f secs' % (after - start,)
            print '    ==> ', nn
        if ctr%100 == 0:
            print '%d done!' % (ctr,)
        nn_pairs.append((lexeme,nn))
    (base, final_ext) = os.path.splitext(nearest_neighbor_file)
    if pos:
        pos_str = '_' + pos
    else:
        pos_str = ''
    full_nearest_neighbor_file = base + pos_str + suffix + ext + final_ext
    if preserve_pos and nearest_neighbor_file:
        if preserve_pos:
            preserve_str = 'pp'
        else:
            preserve_str = ''
        dump_rel_nearest_neighbor_dict (nearest_neighbor_file,preserve_str=preserve_str,pos=pos,suffix=suffix,ext=ext,wt_file=wt_file,top_n=top_n)
    elif nearest_neighbor_file:
        dump_nn_dict('Nearest neighbors','nearest_neighbor_dict', \
                  full_nearest_neighbor_file)
    return nn_pairs

def find_nearest_neighbor (lexeme, pos='', deprel_dict=word_deprel_ct_dict, sim_func='dot_prod',wt_func=None):
    sim_func = getattr(sim_measures,sim_func)
    return __find_nearest_neighbor__ (lexeme, pos, deprel_dict, sim_func,wt_func)

def __find_nearest_neighbor__ (lexeme, pos, deprel_dict, sim_func,wt_func,\
                               external_nn_dict=None,use_sqs=False,\
                               sqs_dict=None, rank_dict=None):
    """
    It is possible for the search for nearest neighbor to have
    no significant result.  For example if the dot prod with
    other lexeme vecs is always 0.  In that case it's okay
    to leave the absurd result, nearest neighbor = ''.

    best_val is the absurd easy-to-beat initial value for
    a sim score. With cosine, it is 0.  With sim scores that range
    from neg infinity to pos infinity this approach to initial
    values cd lead to trouble. In that
    case the correct rewrite is to set best_val initially to None
    and test whether it is not None before testing ">".

    C{preserve_pos}=True means find the nearest neighbor with the same
    pos as C{lexeme}.

    C{pos}=NN means find nearest neighbors that are of pos NN.
    """
    global nearest_neighbor_dict, rel_nearest_neighbor_dict
    if pos and external_nn_dict is not None:
        rel_nearest_neighbor_dict = {}
        rel_nearest_neighbor_dict[pos] = external_nn_dict
        #print '_NN_ dict length: %d' % (len(external_nearest_neighbor_dict),)
    elif external_nn_dict is not None:
        nearest_neighbor_dict = external_nn_dict
        #print '_NN_ dict length: %d' % (len(nearest_neighbor_dict),)
    check_lexeme_type(lexeme, usage_str="find_nearest_neighbor(lexeme,pos='')")
    if pos:
        check_pos_type(pos, usage_str="find_nearest_neighbor(lexeme,pos='')")
    if pos and pos in rel_nearest_neighbor_dict and \
           lexeme in rel_nearest_neighbor_dict[pos]:
        return rel_nearest_neighbor_dict[pos][lexeme]
    if lexeme in nearest_neighbor_dict:
        if pos:
            (word2, pos2) = nearest_neighbor_dict[lexeme]
            if pos == pos2:
                return (word2, pos2)
        else:
            return nearest_neighbor_dict[lexeme]            
    nearest_neighbor = ''
    best_val, comp_rel = sim_measures.sim_func_attributes(sim_func,wt_func)
    this_vec = _word_vec_(lexeme,use_sqs,deprel_dict,sqs_dict)
    if use_sqs:
        sim_func = sim_measures.use_sqs[sim_func]
    if rank_dict:
        lower_rank_bound,upper_rank_bound = get_rank_bounds(lexeme,rank_dict)
        #print 'Bounds: %.4f, %.4f' % (lower_rank_bound, upper_rank_bound)
    else:
        lower_rank_bound = None
        upper_rank_bound = None
    for other_lexeme in deprel_dict:
        (other_word,other_pos) = get_form_pos (other_lexeme)
        if (other_lexeme == lexeme):
            continue
        elif pos and other_pos <> pos:
            continue
        elif lower_rank_bound and upper_rank_bound:
            other_lexeme_rank = rank_dict[other_lexeme]
            if other_lexeme_rank < lower_rank_bound:
                continue
            if other_lexeme_rank > upper_rank_bound:
                continue
        other_vec = _word_vec_(other_lexeme,use_sqs,deprel_dict,sqs_dict)
        new_val = word_vec_sim_score(this_vec, other_vec,sim_func=sim_func)
        # new_val = dot_prod(this_vec, deprel_dict[other_lexeme])
        if comp_rel(new_val,best_val):
            best_val = new_val
            nearest_neighbor = other_lexeme
    nn = (nearest_neighbor, best_val)
    if pos:
        if pos in rel_nearest_neighbor_dict:
            rel_nearest_neighbor_dict[pos][lexeme] = nn
        else:
            this_dict = {}
            this_dict[lexeme] = nn
            rel_nearest_neighbor_dict[pos] = this_dict
    else:
        nearest_neighbor_dict[lexeme] = nn
    return nn

def get_rank_bounds(lexeme, rank_dict):
    return get_unlogged_log_neighborhood(rank_dict[lexeme])

def get_unlogged_log_neighborhood(num,lower_delta=math.e):
    log_num = math.log(num)
    upper_delta_factor = 2* math.e/(log_num+1)
    return (math.exp(log_num - lower_delta), math.exp(log_num + (math.e*upper_delta_factor)))


def merge_word_form_counts(word_deprel_ct_dict, word_ct_dict, lemma_dict):
    """
    Add up the cts for the forms of a lexeme and store in
    separate lexeme ct dicts.
    
    """
    (lexeme_deprel_ct_dict,lexeme_ct_dict) = ({},{})
    for lemma in lemma_dict:
        lexeme = get_form_pos (lemma)
        lexeme_ct_dict[lexeme] = 0
        this_l_dict = {}
        lexeme_deprel_ct_dict[lexeme] = this_l_dict
        for word_form in lemma_dict[lemma]:
            this_f_dict = word_deprel_ct_dict[word_form]
            for deprel in this_f_dict:
                this_l_dict.setdefault(deprel,0)
                this_deprel_ct = this_f_dict[deprel]
                this_l_dict[deprel] += this_deprel_ct
                lexeme_ct_dict[lexeme] += this_deprel_ct
    return (lexeme_deprel_ct_dict,lexeme_ct_dict) 

######################################################################################
######################################################################################
## 
##    N e a r e s t    N e i g h b o r    G a m e s
######################################################################################
######################################################################################

def find_neighborhood (center, satellite, preserve_pos=True, sort=False, sim_func='dot_prod'):
    sim_func = getattr(sim_measures,sim_func)
    __find_neighborhood__(center,satellite,  preserve_pos, sort, sim_func)

def __find_neighborhood__ (center, satellite, preserve_pos, sort, sim_func):
    neighbors = set([])
    T = __sim_score__(center,satellite,sim_func=sim_func)
    (word, pos) = center
    pos_ct = 0
    for (word_i,pos_i) in  word_ct_dict:
        if (word_i,pos_i) == (word,pos):
            continue
        elif preserve_pos and pos_i <> pos:
            continue
        pos_ct += 1
        if __sim_score__(center, (word_i,pos_i), sim_func=sim_func) > T:
            neighbors.add((word_i,pos_i))
    hit_count = len(neighbors)
    if preserve_pos:
        print '%d neighbors were found, %.3f%% of the %ss'  % (hit_count, 100*float(hit_count)/pos_ct, pos)
    if sort:
        sort(list(neighbors))
    return neighbors

#import heapq

# from heapq import heappush, heapreplace, heappop
#  JMG: Finally understood heap implementation better
#   heappushpop is what you want for a best n list.
#   It adds an element x to nbest list L, temporarily making L a list of len n+1,
#   then pops the least element of L.  So if x's score is below the current min, it gets pushed
#   on, then popped off, leaving L unchanged.
#   heapreplace on the other hand ALWAYS replaces the min element of L.
from heapq import heappush, heappushpop, heappop

def find_nearest_n (center,n, preserve_pos=False, pos=None, sim_func='dot_prod',search_space=word_deprel_ct_dict,\
                    deprel_dict=word_deprel_ct_dict,center_vec=None):
    sim_func = getattr(sim_measures,sim_func)
    return __find_nearest_n__ (center, n, preserve_pos, pos, sim_func, search_space, deprel_dict, center_vec)
                    
def __find_nearest_n__ (center, n, preserve_pos, pos, sim_func, search_space, deprel_dict, center_vec):
    """
    Find nearest C{n} neighbors of C{center} in C{search_space}. C{search_space} is any container with a set of objects
    whose distance can be measured using C{sim_func}.

    C{deprel_dict} contains vectors for C{center} and objects in C{search_space} that can be used for the
    C{sim_func} computation.

    C{center_vec} is used when the arg is a centroid and actually has no entry in the deprel_dict.
    Some string pair must still be passed in to stand in as C{center}, the name for the fictional lex entry being used.
    """
    neighbors = []
    if n == 0:
        return neighbors
    heap_len = 0
    # heap_min = 0
    if pos:
        this_pos = pos
    elif preserve_pos:
        (word, this_pos) = center
    else:
        this_pos = None
    if not center_vec:
        center_vec = deprel_dict[center]
    for lexeme in  search_space:
        if lexeme == center:
            continue
        elif this_pos:
            (word_i,pos_i) = lexeme
            if pos_i <> this_pos:
                continue
        score = efficient_sim_score(center_vec, lexeme,sim_func=sim_func,deprel_dict=deprel_dict)
        if heap_len < n:
            heappush(neighbors,(score, lexeme))
            heap_len += 1
            # if score > heap_min:
            #    pass
            # else:
            #    heap_min = score
        else:
            #if score > heap_min:
            #    ## Using replace rather than pushpop  because neighbors cant be empty
            #    heapreplace(neighbors,(score,lexeme))
            #    ## Reset min.
            #    (heap_min, min_lex) = heappop(neighbors)
            #    heappush(neighbors,(heap_min,min_lex))
            heappushpop(neighbors, (score,lexeme))
    return [heappop(neighbors) for i in range(n)]


def get_rank_results (dict_dir, rank_dict,dependency_train_file):
    rank_dict_dict = {}
    rank_results = {}
    for f in os.listdir(dict_dir):
        (base,ext) = os.path.splitext(f)
        ## Strip off long filename prefix to get to the meat
        prop = base[24:]
        (nn_rank_dict, avg, neighbor_freq, neighbor_dict) = \
                       get_nn_ranks(os.path.join(dict_dir,f),\
                                    rank_dict=rank_dict,\
                                    filename=dependency_train_file,
                                    prop=prop)
        if nn_rank_dict:
            rank_dict_dict[prop] = (nn_rank_dict,neighbor_dict)
            rank_results[prop] = (avg,neighbor_freq)
    il = rank_results.items()
    il.sort(reverse=True,key=lambda x: x[1][0])
    for (k,(v,nf)) in il:
        print '%-30s  & %-6d [%.3f]\\\\' % (k,v,nf)
    return (rank_dict_dict, rank_dict, rank_results)
            


def get_nn_ranks (ext_nn_dict_file,rank_dict = {}, filename='', prop=''):
    assert rank_dict or filename, 'Must give info source for ranks, either a rank dict or a dependency db count file'
    if not rank_dict:
        print 'No rank dict'
        make_ct_dict_fill_rank_dict(filename,rank_dict)
    ext_nn_dict = {}
    nn_rank_dict = {}
    neighbor_dict = {}
    ctr = 0
    rank_tot = 0.0
    read_nn_file(ext_nn_dict_file, nn_dict = ext_nn_dict)
    nn_set = set([])
    for (lex,(nn,score)) in ext_nn_dict.iteritems():
        ctr += 1
        try:
            nn_r = rank_dict[nn]
        except KeyError:
            print 'Key Error: %s' % (nn,)
            continue
        nn_set.add(nn)
        rank_tot += nn_r
        nn_rank_dict[lex] = (nn, nn_r)
        if nn in neighbor_dict:
            neighbor_dict[nn].append(lex)
        else:
            neighbor_dict[nn] = [lex]
    try:
        avg = rank_tot/ctr
    except ZeroDivisionError:
        return ({}, None)
    try:
        neighbor_freq = float(len(nn_set))/ctr
    except ZeroDivisionError:
        return ({}, None)
    #print '%s &  %.3f \\\\' % (prop, avg,)
    return (nn_rank_dict, avg, neighbor_freq, neighbor_dict)
        
def make_ct_dict_fill_rank_dict (filename, rank_dict):
    il = dep_file_utils.get_sorted_word_counts(filename)
    ct_dict = {}
    ctr = 0
    for (k,v) in il:
        ctr += 1
        rank_dict[k] = ctr
        ct_dict[k] = v
    return ct_dict
    

def compare_nn_dicts (ext_nn_dict_file,pos='NN', reference_dict=None,\
                      reference_dict_file=None):
    """
    Check how different the current nn dict is from some pre-made competitor. or how
    different a given reference dict is from an external file. Or compare two
    nn_dict files.

    Count and print diffs, ignoring scores.

    Although one cant specify both C{reference_dict} and C{reference_dict_file}, one can specify
    a file and pass in an empty dict as the C{reference_dict} parameter. This dictionary is
    then filled with the contents of C{reference_dict_file}.
    """
    global rel_nearest_neighbor_dict
    assert not (reference_dict and reference_dict_file), "There can be only one reference dict.\nYou cannot supply both reference dict and reference dict file arguments."
    ####  reference_dict
    if reference_dict_file:
        # Allow an empty ref dict arg to be passed in and used.
        if not reference_dict == {}:
            reference_dict = {}
        read_nn_file(reference_dict_file, nn_dict = reference_dict)
    elif reference_dict == None:
        reference_dict = rel_nearest_neighbor_dict[pos]
    ####  ext_nn_dict
    ext_nn_dict = {}
    read_nn_file(ext_nn_dict_file, nn_dict = ext_nn_dict)
    _compare_nn_dicts_ (reference_dict, ext_nn_dict,print_diffs=True)


def _compare_nn_dicts_ (reference_dict, ext_nn_dict,key=None, print_diffs=False):
    ### Count diffs
    ctr = 0
    il = reference_dict.items()
    for (k,v) in il:
       try:
               if reference_dict[k][0] <> ext_nn_dict[k][0]:
                      ctr += 1
       except KeyError:
           # Allow for the possibility of vocab divergence.  We compare NNs
           # only for SHARED vocab.
           continue
    print '%d differences were found!' % (ctr,)
    if print_diffs:
        print_nn_dict_diffs (reference_dict, ext_nn_dict, key=key)

def print_nn_dict_diffs (reference_dict, ext_nn_dict, key=None):
    format_str = '%-20s\t%-20s\t%-20s'
    print
    print format_str % ('Test', 'Reference', 'External')
    print '=' * len(format_str % ('Test', 'Reference', 'External'))
    il = reference_dict.items()
    if key:
        il.sort(key=key)
    for (k,v) in il:
           try:
               if reference_dict[k][0] <> ext_nn_dict[k][0]:
                      print format_str % (k[0], reference_dict[k][0][0], ext_nn_dict[k][0][0])
           except KeyError:
                  continue

   
def lemma_update(lexeme,lemma_dict,wnl,no_proper_names=True):
    (form, bnc_pos) = lexeme
    if no_proper_names and bnc_pos in ['NP', 'NPS']:
        # save room in lemma_dict
        return
    wn_pos = bnc_pos_to_wn_pos(bnc_pos)
    if wn_pos:
        lemma_list = wnl(form,wn_pos)
        for lemma in lemma_list:
            lemma = lemma + '.' + wn_pos
            if lemma in lemma_dict:
                lemma_dict[lemma].add(lexeme)
            else:
                lemma_dict[lemma] = set([lexeme])


######### Tests, Utilities

# Computed for MacOSX 10.6.
# May break on another platform

min_float = 1e-255

def float_gt_eq (l,r):
    """
    A better def?
    return l > r or not r > l
    """
    return l > r or ((abs(r - l)) < min_float)



def sim_score (lexeme1, lexeme2,sim_func='dot_prod',deprel_dict=word_deprel_ct_dict):
    """
    Currently implemented vals for sim_func: dot_prod, jaccard, and IRad.
    cosine is implemented by doing L2 normalization of all the vectors
    and then using dot_prod as the sim_func.  IRad assumes the
    vectors have probs representing for each feat f its conditional prob: p(f | lexeme).
    Practically this means the counts have been L1-normalized.
    """
    sim_func = getattr(sim_measures,sim_func)
    return sim_func(deprel_dict[lexeme1], deprel_dict[lexeme2])

def __sim_score__ (lexeme1, lexeme2,sim_func,deprel_dict=word_deprel_ct_dict):
    """
    Currently implemented vals for sim_func: dot_prod, jaccard, lin, dice_dagger, and IRad.
    cosine is implemented by doing L2 normalization of all the vectors
    and then using dot_prod as the sim_func.  IRad assumes the
    vectors have probs representing for each feat f its conditional prob: p(f | lexeme).
    Practically this means the counts have been L1-normalized.
    """
    return sim_func(deprel_dict[lexeme1], deprel_dict[lexeme2])

def efficient_sim_score (vec1, lexeme2,sim_func,deprel_dict=word_deprel_ct_dict):

    return sim_func(vec1, deprel_dict[lexeme2])

def word_vec_sim_score (vec1, vec2,sim_func):

    return sim_func(vec1, vec2)


def _word_vec_(lexeme,use_sqs,deprel_dict,sqs_dict):
    this_vec = deprel_dict[lexeme]
    if use_sqs:
        return (this_vec,sqs_dict[lexeme])
    else:
        return this_vec


def pp_word_vec (lex,deprel_dict=word_deprel_ct_dict):
    vec =  _word_vec_(lex,deprel_dict)
    for (feat,val) in vec:
        print 


#################################################################
#################################################################
####  Scoring utilities
#################################################################
#################################################################



def score_nn_dict (gold_file, output_file=None):
    """
    Use a gold file to score the CURRENT nearest_neighbor_dict.
    Make noisy report.
    
    Output to file if requested.
    """
    print 'Reading gold file %s' % (gold_file,)
    (good,bad) = read_gold_nn_file (gold_file)
    print 'Gold file %s read.' % (gold_file,)
    (good_d, bad_d, unscored) = score_nn_dict_helper(good, bad)
    good_ct = len(good_d)
    total = good_ct + len(bad_d)
    report_str = '%d total pairs: %d were scored, %d were good  (%.2f per cent)' % (len(nearest_neighbor_dict), total, good_ct, (100 *float(good_ct))/total)
    print report_str
    if output_file:
        dump_score_dict(good_d,bad_d,unscored,report_str,output_file)


def score_nn_dict_helper(good, bad):
    """
    Close analogue of C{score_pairs} which takes
    C{good} and C{bad} dictionaries as input
    and uses it to sort pairs from the
    current C{nearest_neighbor_dict}.
    """
    (good_d, bad_d, unscored) = ({ }, { }, { })
    for l1 in nearest_neighbor_dict:
        (l2, sim) = nearest_neighbor_dict[l1]
        if (l1,l2) in good:
            good_d[l1] = (l2,sim)
        elif (l1,l2) in bad:
            bad_d[l1] = (l2,sim)
        else:
            unscored[l1] = (l2,sim)
    return (good_d, bad_d, unscored) 


def score_pairs(pairs, good, bad):
    """
    Utility function for splitting a list of nn pairs C{pairs}
    into C{bad} C{bad} and C{unscored} given separate
    lists of good and bad pairs.  used only in
    making mnn reports to sort the output of
    find_mutual_nearsets_neighbors when there's
    info from a gold file.
    """
    (good_l, bad_l, unscored) = ([ ], [ ], [ ])
    for (l1,l2) in pairs:
        if (l1,l2) in good:
            good_l.append((l1,l2))
        elif (l1,l2) in bad:
            bad_l.append((l1,l2))
        else:
            unscored.append((l1,l2))
    return (good_l, bad_l, unscored) 



#################################################################
#################################################################
###  Mutual nearest Neighbor reports  ###
#################################################################
#################################################################

def make_mutual_nearest_neighbor_report (mnn_file, gold_file=None, mnn_l=None,\
                                         suffix='', ext=''):
    """
    C{mnn_file} is output file to dump report to.
    
    Options::
      -- Get mutual_neighbors_list precomputed C{mnn_l} and report
      on it. Or compute it here and report opn it.
      -- Get gold_file with human-evals on the pairs in nearest
      neighbors dict and use it in report.
      
    Assumption: C{nearest_neighbor_dict} exists. Plus, whatever the assumptions of
    C{find_mutual_neaerest_neighbors} are about C{nearest_neighbor_dict}.
    """
    (base,final_ext) = os.path.splitext(mnn_file)
    full_mnn_file = base + suffix + ext + final_ext
    if not (suffix and ext):
        print 'Warning: No suffix and ext given.  Writing report to %s' % (full_mnn_file,) 
    if gold_file:
        (good,bad) = read_gold_nn_file (gold_file)
    if mnn_l:
        res = mnn_l
    else:
        res = find_mutual_nearest_neighbors()
    dump_nn_dict('Nearest neighbors','nearest_neighbor_dict', full_mnn_file)
    if gold_file:
        (good_l, bad_l, unscored) = score_pairs(res,good,bad)
        good_ct = len(good_l)
        num_scored = good_ct + len(bad_l)
        ofh = open(full_mnn_file,'a')
        print >> ofh
        score_report_str = 'Out of %d Mutual nearest neighbors found, %d were scored, %d were good (%.2f)'
        print >> ofh, score_report_str % (len(res), num_scored, good_ct, float(good_ct)/num_scored)
        ofh.close()
        append_pair_list_to_file('Good Mutual nearest neighbors', good_l, mnn_file)
        append_pair_list_to_file('Bad Mutual nearest neighbors', bad_l, mnn_file)
        append_pair_list_to_file('Unscored Mutual nearest neighbors', unscored, mnn_file)
    else:
        append_pair_list_to_file('Mutual nearest neighbors', res, mnn_file)

def dump_score_dict (good_d,bad_d,unscored,report_str,output_file):
    """
    good_d, bad_d, unscored are dictionaries containing good, bad,
    and unscored nearest neighbor pairs as keys. report_str contains some
    stats. Output_file is a filepath represented as a string.
    """
    ofh = open(output_file,'w')
    if len(unscored) > 0:
        print >> ofh, 'Partially scored gold standard file for Nearest neighbor pairs.'
        print >> ofh, 'Gold standard file for Nearest neighbor pairs.'
    print >> ofh, report_str
    print >> ofh
    for k in good_d:
        print >> ofh, k, good_d[k], 1
    for k in bad_d:
        print >> ofh, k, bad_d[k], 0
    for k in unscored:
        print >> ofh, k, unscored[k]
    ofh.close()

#################################################################
#################################################################
###  End Mutual nearest Neighbor reports  ###
#################################################################
#################################################################


#################################################################
#################################################################
## Utilities
#################################################################
#################################################################


def reset_rel_nearest_neighbor(pos):
    old_dict = rel_nearest_neighbor_dict[pos]
    rel_nearest_neighbor_dict[pos] = {}
    return old_dict

results_dict = {}

def run_nearest_neighbor_experiment (dependency_train_file, nearest_neighbor_file,\
                                     wt_func_nickname, sim_func_list, pos, verbose=False, top_n=500,wt_func=None):
    """
    Hold C{wt_func} constant because a particular wt function model has just been loaded.  Run
    through a given set of sim functions (C{sim_func_list}) computing nearest neighbors for some fixed top_n list.
    Store copies of all nn dicts computed in the global C{results_dict}.  But output each nn_dict to a file too using
    the usual file naming conventions.

    Example::
       -- run_nearest_neighbor_experiment (wt_func='t_test_no_norm', top_n=(500,1000), ['dot_prod', 'lin_no_neg','dice_dagger'])
    """
    global results_dict
    results_dict = {}
    for sim_func in sim_func_list:
        find_nearest_neighbors (dependency_train_file, nearest_neighbor_file=nearest_neighbor_file, \
                                top_n=top_n, preserve_pos=True, sim_func=sim_func, \
                                pos=pos, ext='_' + wt_func_nickname, verbose=verbose,wt_func=wt_func)
        results_dict[wt_func_nickname +'_' + sim_func] = reset_rel_nearest_neighbor('NN')



#################################################################
#################################################################
## read Utilities
#################################################################
#################################################################

import re
## Match ((word,pos), ((neighbor,pos), score))
nn_dd_line = re.compile(r'(\(.+\))\s*\((\(.+\))\s*,\s*(.+)\)')

def read_nn_file (nn_file, suffix='',ext='',nn_dict=None):
    (base,finalext) = os.path.splitext(nn_file)
    fh = open(base + suffix + ext + finalext,'r')
    match_found = False
    if nn_dict is None:
        nn_dict = nearest_neighbor_dict
    for line in fh:
        line = line.strip()
        match = nn_dd_line.match(line)
        if match:
            match_found = True
            groups = match.groups()
            nn_dict[eval(groups[0])] = (eval(groups[1]),float(groups[2]))
    if not match_found:
        print 'Not a nearest neighbor file!'
    fh.close()
    return nn_dict

class NNError (Exception):
    pass

def read_nn_file_safe (nn_file, suffix='',ext='',nn_dict=None):
    """
    Raises an Exception if this is not an NN file.
    """
    (base,finalext) = os.path.splitext(nn_file)
    final_filename = base + suffix + ext + finalext
    match_found = False
    if nn_dict is None:
        nn_dict = {}
    with open(final_filename,'r') as fh:
        for line in fh:
            line = line.strip()
            match = nn_dd_line.match(line)
            if match:
                match_found = True
                groups = match.groups()
                nn_dict[eval(groups[0])] = (eval(groups[1]),float(groups[2]))
    if not match_found:
        raise NNError, '%s is not a nearest neighbor file!' % (final_filename,)
    return nn_dict

def read_nn_file_to_list_safe (nn_file, suffix='',ext=''):
    """
    Read an nn file into alist preserving rank order in the nn file.
    
    Raises an Exception if this is not an NN file.
    """
    (base,finalext) = os.path.splitext(nn_file)
    final_filename = base + suffix + ext + finalext
    match_found = False
    nn_list = []
    with open(final_filename,'r') as fh:
        for line in fh:
            line = line.strip()
            match = nn_dd_line.match(line)
            if match:
                match_found = True
                groups = match.groups()
                nn_list.append((eval(groups[0]), (eval(groups[1]),float(groups[2]))))
    if not match_found:
        raise NNError, '%s is not a nearest neighbor file!' % (final_filename,)
    return nn_list

def read_gold_nn_file (nn_file):
    fh = open(nn_file,'r')
    good_pairs, bad_pairs = ({}, {})
    match_found = False
    for line in fh:
        line = line.strip()
        match = nn_dd_line.match(line)
        if match:
            match_found = True
            score = line.split()[-1]
            try:
                score = int(score)
            except ValueError:
                print 'Not a gold standard NN file. Unscored?'
                break
            groups = match.groups()
            if score > 0:
                good_pairs[eval(groups[0]),eval(groups[1])] = float(groups[2])
                good_pairs[eval(groups[1]), eval(groups[0])] = float(groups[2])
            else:
                bad_pairs[eval(groups[0]) , eval(groups[1])] = float(groups[2])
                bad_pairs[eval(groups[1]), eval(groups[0])]  = float(groups[2])
    if not match_found:
        print 'Not a nearest neighbor file!'
    fh.close()
    good_ct = len(good_pairs)/2
    bad_ct = len(bad_pairs)/2
    total = good_ct + bad_ct
    print 'Out of %d total pairs %d were scored good  (%.2f per cent)' % (total, good_ct, (100 *float(good_ct))/total)
    return (good_pairs,bad_pairs)

def read_sqs_file (dependency_file,suffix,ext):
    full_dependency_file = \
           make_mi_filename(dependency_file, '', False, suffix, ext)
    (filepath,ext) = os.path.splitext(full_dependency_file)
    sqs_file = filepath +'.sqs'
    print 'Reading sqs fle %s' % (sqs_file,)
    fh = open(sqs_file,'r')
    sqs_dict = {}
    for line in fh:
        try:
            (lexeme, sum_sqs) = \
                     dep_file_utils.parse_ct_file_line(line, sqs_file, ct_type=float)
            sqs_dict[lexeme] = sum_sqs
        except TypeError,e:
            raise e
    return sqs_dict


#################################################################
#################################################################
## End read Utilities
#################################################################
#################################################################



#################################################################
#################################################################
## write Utilities
#################################################################
#################################################################

pruning_dict = \
{True: '_pruned',
 False: ''}

def make_mi_filename(dependency_file, prefix, prune, suffix, ext):
    if prefix:
        (path,filename) = os.path.split(dependency_file)
        dependency_file = os.path.join(path,prefix + filename)
    (base, final_ext) = os.path.splitext(dependency_file)
    pruning_affix = pruning_dict[prune]
    return base + pruning_affix + suffix + ext + final_ext

def output_dependency_dict(word_deprel_ct_dict, dependency_file, \
                           suffix='',  ext='', prefix='', count_file=False,
                           prune=False,new_wt_file=None,write_dir='',\
                           threshhold=8,store_sqs=False,sqs_dict=None,\
                           word_ct_dict=None):
    """
    default threshhold = 8.  We want to prune wds with wd <= 5.  Min
    feat ct = 5, on average, 1.4 is ration of featct to token ct.
    5 * 1.4 = 7.  Add 1 to use < threshhold test to find killables.
    

    Slightly more readable format than C{dump_nn_dict} intended
    specifically for the original dependency db, with
    no descriptors at top of file.

    With C{count_file} = False, is used to write a the mi_file/prob/pmi after reading
    in the raw word form counts file, and computing the relevant floats.

    With C{count_file} = True, is used to write a true lexeme train file with counts after reading
    in the raw word form counts file.
    """
    # This global is reset.  Why?
    global wt_file
    assert not (prune and not word_ct_dict), 'Word_ct_dict must be supplied to prune'
    ctr = 0
    prune_ctr = 0
    pruned_cells = 0

    if new_wt_file:
        full_dependency_file = os.path.join(write_dir, new_wt_file)
    else:
        full_dependency_file = make_mi_filename(dependency_file, prefix, prune, suffix, ext)
    if store_sqs:
        (filepath,ext) = os.path.splitext(full_dependency_file)
        sfh = open(filepath +'.sqs','w')
    ofh = open(full_dependency_file,'w')
    for lexeme in word_deprel_ct_dict:
        this_dict = word_deprel_ct_dict[lexeme]
        if prune:
            size = word_ct_dict[lexeme] #sum(vals)
            if size < threshhold:
                prune_ctr += 1
                pruned_cells += len(this_dict.values())
                continue
        ctr += 1
        print >> ofh, make_lexeme_str(lexeme),
        keylist = this_dict.keys()
        keylist.sort()
        for deprel in keylist:
            print >> ofh, make_deprel_str(deprel,this_dict,count_file),
        print >> ofh
        if store_sqs:
            print >> sfh, '%s %.6f' % (lexeme, sqs_dict[lexeme])
    ofh.close()
    if store_sqs:
        sfh.close()
    if not count_file:
        wt_file = full_dependency_file
    print '%d cells pruned by word pruning' % pruned_cells
    print '%d lexical items were saved; %d were pruned; squares saved: %s' % (ctr,prune_ctr,store_sqs)

def make_lexeme_str (lexeme):
    return '%s=%s' % get_form_pos(lexeme)

def make_deprel_str (deprel,this_dict,count_file):
    if count_file:
        return '%d:%d' % (deprel,this_dict[deprel])
    else:
        return '%d:%.8f' % (deprel,this_dict[deprel])

def get_form_pos (lexeme):
    try:
        (word,pos) = lexeme
    except ValueError:
        # In lemma form
        try:
            (word,pos) = lexeme.split('.')
        except ValueError:
            # Extra '.'
            # print 'Warning: %s is a lexeme.' % (lexeme,)
            formlist = lexeme.split('.')
            (word,pos) = ('.'.join(formlist[:-1]),formlist[-1])
    return (word,pos)

def dump_nn_dict (description, dd_name, dd_file, env = {}):
    """
    dd_name is a string which evals to a dictionary, which allows it
    to be included in the first line file descriptor to specify the attribute
    name the dictionary shd have when read back in.

    Used for dumping nearest neighbor dicts, say, for an update after running
    C{find_mutual_nearest_neighbors} has changed the C{nearest_neighbor_dict}.

    This can also be used to create nn_files that are to be annotated with
    scores (1= gd, 0 = bd).
    """
    if not type(dd_name) == str:
        print 'Usage: dump_nn_dict <dictionary description> <dictionary-name-as-string> <file-name>'
        dd_name = ''
    elif env:
        dd = env[dd_name]
    else:
        dd = eval(dd_name)
    ofh = dep_file_utils.open_and_make_backup(dd_file)
    print >> ofh, description + ': %s' % (dd_name,)
    print >> ofh
    for k in dd:
        print >> ofh, k, dd[k]
    ofh.close()

def dump_partial_rel_nn_dict (dd_file,pos):
    pos_dict = rel_nearest_neighbor[pos]
    dump_dict(dd_file,pos_dict)
    
def dump_dict (dd_file, dd):
    """
    Just dump the damned dict, no name games.
    """
    ofh = open(dd_file,'w')
    for k in dd:
        print >> ofh, k, dd[k]
    ofh.close()
    
    

def dump_rel_nearest_neighbor_dict (nearest_neighbor_file, preserve_str='',\
                                    pos='', suffix='', ext='', wt_file=None, top_n=0):
    """
    suffix = simfn
    ext = space
    """
    global rel_nearest_neighbor_dict
    base, final_ext = os.path.splitext(nearest_neighbor_file)
    prefix = '_'.join(['rel',preserve_str,base])
    full_fn = make_nn_file_name(pos=pos, simfn=suffix[1:], space=ext[1:], \
                                prefix=prefix, dict_ext=final_ext, top_n=top_n)
    if os.path.isfile(full_fn):
        new_name = dep_file_utils.find_unique_backup_name(full_fn)
        print 'Renaming %s to be %s' % (full_fn, new_name)
        os.rename(full_fn, new_name)
    fh = open(full_fn,'w')
    if pos:
        pos_str = 'Pos %s: ' % pos
        to_do = [pos]
    else:
        pos_str = ''
        to_do = rel_nearest_neighbor_dict.keys()
    print >> fh, '%sNearest neighbors of same pos: %s' % (pos_str, 'rel_nearest_neighbor_dict')
    print >> fh
    for k in to_do:
        this_dict = rel_nearest_neighbor_dict[k]
        if wt_file:
            print >> fh, 'wt file: %s' % (wt_file,)
        print >> fh, '%d most frequent %s' % (len(this_dict), k)
        for k in this_dict:
            print >> fh, k, this_dict[k]
        print >> fh
    fh.close()

def append_pair_list_to_file (name, ll, ll_file):
    ofh = open(ll_file,'a')
    print >> ofh
    print >> ofh, name
    print >> ofh
    for (x,y) in ll:
        print >> ofh, x, y
    ofh.close()

###  Simple(minded) type checking

def check_lexeme_type(lexeme, usage_str=''):
    try:
        assert(type(lexeme) == tuple and len(lexeme) == 2)
        assert(type(lexeme[0]) == str)
        assert(type(lexeme[1]) == str)
    except AssertionError:
        print 'Usage: %s' % (usage_str,)
        raise Exception, 'lexeme %s is not a lexeme.' % (lexeme,)
    
def check_pos_type(pos, usage_str=''):
    try:
        assert(type(pos) == str)
    except AssertionError:
        print 'Usage: %s' % (usage_str,)
        raise Exception, 'Pos %s is not a part of speech.' % (pos,)
   

def make_nn_file_name(pos, simfn, space, prefix='rel_pp_nearest_neighbor', dict_ext='.dict', top_n=0):
    """
    Assumption:: C{nn_dict}s in the dep_db directory have been named according to these
    conventions, fixed prefix plus a pos, simfn, and space component. Passing in a frequency range
    tuple as C{top_n} results in a frequency range infix in the filename.
    """
    if top_n:
        if type(top_n) == tuple:
            top_n = '%s_%s' % (top_n[0],top_n[1])
        else:
            top_n = ''
    else:
        top_n = ''
    parts_list = [x for x in [prefix,pos,top_n,simfn,space] if x]
    return '_'.join(parts_list) + dict_ext

def make_neighborhood_graph (lexeme, n=5, dot_file=False, preserve_pos=False, sim_func='dot_prod', \
                             search_space=word_deprel_ct_dict, center_vec=None):
    (word, pos) = lexeme
    ngraph = { }
    nn_list = find_nearest_n(lexeme, n, preserve_pos = preserve_pos, sim_func=sim_func, search_space=search_space,\
                             center_vec=center_vec)
    (score, nn) = nn_list[-1]
    word = mk_word_str(word, pos)
    process_nn_link (ngraph, word, nn[0], nn[1],preserve_pos=preserve_pos,sim_func=sim_func)
    for (score, (other_word, other_pos)) in nn_list[:-1]:
        process_nn_link (ngraph, word, other_word, other_pos, n,preserve_pos=preserve_pos,sim_func=sim_func)
    if dot_file:
        make_dot_file(ngraph, word, center=word, preserve_pos=preserve_pos)
    return ngraph

def process_nn_link (ngraph, word, other_word, other_pos,n=0,preserve_pos=False,sim_func='dot_prod'):
    """
    DONT make this recursive.
    """
    t_other_word = mk_word_str(other_word,other_pos)
    if n:
        add_graph_link(ngraph, word, (n, t_other_word))
    else:
        add_graph_link(ngraph, word, t_other_word)
    if preserve_pos:
        next_pos = other_pos
    else:
        next_pos = ''
    ((nn_word, nn_pos),other_score) = \
               find_nearest_neighbor((other_word,other_pos),pos=next_pos,sim_func=sim_func)
    add_graph_link(ngraph,t_other_word, mk_word_str(nn_word,nn_pos))
    
def mk_word_str (w,p):
    """
    Much that cd be done here.
    
    For now, deal with homophony of finite and nonfinite
    verbs by renaming finites in graph
    """
    if p == 'VBP':
        return w + '_f'
    else:
        return w

def add_graph_link (g,w, nn):
    if w in g:
        g[w].append(nn)
    else:
        g[w] = [nn]

# Args: color, fillcolor
graph_template = \
"""
digraph G {
        node [color=%s, fillcolor=%s,style="filled,setlinewidth(2)"];
"""

# Arc where to-node is one of n nearest neighbors of from-node
# Args: from-node to-node
n_arc_template = '       %s -> %s ;'
## Arc for true nearest Neighbors
arc_template = '       %s -> %s [color = "red"] ;'
end_graph_template = '}'
center_template = '        %s [color="black",fillcolor="sandybrown"] ;'

def make_dot_file (link_dict,filename,color='black',fillcolor='salmon2', \
                   center=None,preserve_pos=False):
    line_list = [graph_template % (color, fillcolor)]
    if center:
        line_list.append(center_template % (center,))
    for (key, vallist) in link_dict.items():
        for val in vallist:
            if type(val) == tuple:
                line_list.append(n_arc_template % (key, val[1]))
            else:
                line_list.append(arc_template % (key, val))
    line_list.append(end_graph_template)
    if preserve_pos:
        filename = filename + '_pp'
    fh = open(filename + '.dot', 'w')
    print >> fh, '\n'.join(line_list)
    fh.close()


def count_words (pos, word_deprel_ct_dict=word_deprel_ct_dict):
    """
    A way of sorting the words from a ct deprel dict, based on deprel ct.
        il = count_words('NN', word_deprel_ct_dict)
    """
    ct_dict = {}
    for (word, other_pos) in word_deprel_ct_dict:
        if pos == other_pos:
            vec = word_deprel_ct_dict[(word,other_pos)]
            res = 0
            for (k,v) in vec.iteritems():
                res += v
            ct_dict[(word,other_pos)] = res
    ct_dict_items = ct_dict.items()
    ct_dict_items.sort(reverse=True,key=lambda x: x[1])
    return ct_dict_items


class Timer():
   """
   From  http://mrooney.blogspot.com/2009/07/simple-timing-of-python-code.html

   An example::

   with Timer():
       time.sleep(1.5)
   """
   def __enter__(self): self.start = time.time()
   def __exit__(self, *args): print time.time() - self.start

################
### Semi retired
###############


####################################
####  PRETTY MUCH RETIRED    ####### 
####################################   

def find_nearest_lexeme_neighbors (dep_file, lexeme_ct_dict, lexeme_deprel_ct_dict, \
                                   nearest_neighbor_file=None, top_n=0, pos='',\
                                   suffix='', ext='', prefix='lemma_'):
    """
    If C{top_n} and C{pos}, get n most frequent words with part of speech C{pos}.
    """
    vocab = lexeme_ct_dict.items()
    if top_n:
        vocab.sort(key=lambda x:x[1],reverse=True)
    ctr = 0
    for (lexeme, ct) in vocab:
        if ctr > top_n:
            break
        (word, l_pos) = get_form_pos (lexeme)
        if pos and l_pos <> pos:
            continue
        ctr += 1
        find_nearest_neighbor(lexeme, pos=l_pos,deprel_dict=lexeme_deprel_ct_dict)
        if ctr%100 == 0:
            print '%d done!' % (ctr,)
    if nearest_neighbor_file:
        if prefix:
            (base, final_ext) = os.path.splitext(nearest_neighbor_file)
            (path,filename) = os.path.split(base)
            nearest_neighbor_file = os.path.join(path,prefix + filename + final_ext)
        dump_rel_nearest_neighbor_dict (nearest_neighbor_file,preserve_str='pp',pos=pos,suffix=suffix,ext=ext)



def find_mutual_nearest_neighbors (nn_dict={ }):
    """
    Assumption: C{nearest_neighbor_dict} has been populated
    using C{find_nearest_neighbors}, at least for the top n
    vocab words.  We dont however assume that the nearest neighbor
    of a top-n word is also a top-n-word.  So we allow for the possibility
    that the nearest neighbor of a nearest neighbor might yet need
    to be computed.

    nn_dict may be the absolute C{nearest_neighbor_dict} (default)
    or it may be a passed in  nn_dict, usually one of the sub dicts of
    the rel_nearest_neighbor dict.
    """
    print 'Finding mutual nearest neighbors'
    global mutual_nearest_neighbors
    done = [ ]
    if not nn_dict:
        nn_dict = nearest_neighbor_dict
    for lexeme in nn_dict.keys():
        if lexeme in done:
            continue
        (nn, sim_score) = nearest_neighbor_dict[lexeme]
        try:
            (nn2, sim_score2) = nearest_neighbor_dict[nn]
        except KeyError:
            (nn2, sim_score2) = find_nearest_neighbor(nn)
        if lexeme == nn2:
            mutual_nearest_neighbors.append((lexeme,nn))
            done.append(nn)
    print 'Mutual nearest neighbors found'
    return mutual_nearest_neighbors

def make_attractor_lists(sim_systems,rank_dict_dict):
    attractor_lists = {}
    for sim_s in sim_systems:
        il = rank_dict_dict[sim_s][1].items()
        attractor_lists[sim_s] = il
        il.sort(reverse=True,key=lambda x: len(x[1]))
    return attractor_lists

def print_attractors (sim_sys, attractor_lists, top_n=20):
    def clean_att_l (att_l):
        att_l = [l[0] for l in att_l]
        return ' '.join(att_l)
    
    print sim_sys
    for (att,att_l) in attractor_lists[sim_sys][:top_n]:
        print '%-20s   %s' % (att[0], clean_att_l(att_l))


def reset_dep_db_globals():
    global word_deprel_ct_dict, deprel_ct_dict, word_ct_dict, lemma_dict, lexeme_deprel_ct_dict, lexeme_ct_dict
    
    (word_deprel_ct_dict, deprel_ct_dict, word_ct_dict, lemma_dict, lexeme_deprel_ct_dict, lexeme_ct_dict) = ({ }, { }, { }, { },{ },{ })
    
if __name__ == '__main__':
        
    import os.path

    home = os.environ['HOME']
    #home = '/Users/gawron/'
    dependency_db_dir = os.path.join(home,'Desktop','SWIDSAS','dependency_DB')
    os.chdir(dependency_db_dir)
    # This is the orginal big DB made from Malt + BNC
    #dep_db_dir = '../dep_data'
    #dep_db_dir = '../dep_data_extra'
    ## This is the 2nd big one. Stanford + BNC
    dep_db_dir = os.path.join(home, 'ext','corpus', 'bnc_stanford_dep_db_test')
    #parse_domain = 'dependency_DB'
    parse_domain = 'bnc_stanford'
    #parse_domain = 'bnc_xml_DB'
    #pruned = '_pruned_nonneg'
    #pruned = '_pruned'
    pruned=''

    ## Assumed to exist
    dependency_keys_file = os.path.join(dep_db_dir,'%s.keys' % parse_domain)
    dependency_train_file = os.path.join(dep_db_dir, '%s%s.train' % (parse_domain,pruned))
    #parse_domain = 'vaccine'
    #dependency_db_dir = os.path.join(home,'Desktop', 'SWIDSAS','dependency_DB','clustering','vaccine')
    #dependency_db_dir = \
    #    os.path.join(home,'Desktop', 'SWIDSAS','models','models')
    #dependency_db_dir = \
    #     os.path.join(home,'Ranking','vaccine','nb_vaccine','models','models')
    #dependency_train_file = 'neg_dep_db_pruned_nonneg.train'
    #dependency_train_file = 'neg_dep_db.train'
    #dependency_train_file = 'dep_db_pruned_nonneg.train'
    #dependency_train_file = 'dep_db.train'
    #dependency_train_file = os.path.abspath(os.path.join(dependency_db_dir,dependency_train_file))

    #dependency_keys_file = 'dep_db.keys'
    #dependency_keys_file = 'neg_dep_db.keys'
    dependency_keys_file = os.path.abspath(os.path.join(dependency_db_dir,dependency_keys_file))
    # Using pruned model, wd_thresshold = 6
    # dependency_train_file = os.path.join(dep_db_dir, 'dependency_DB_pruned.train')
    # dependency_keys_file = os.path.join(dep_db_dir, 'bnc_xml_DB.keys')
    # dependency_train_file = os.path.join(dep_db_dir, 'bnc_xml_DB.train')
    #dependency_keys_file = os.path.join(dep_db_dir, 'militant_data_DB.keys')
    #dependency_train_file = os.path.join(dep_db_dir, 'militant_data_DB.train')
    #dependency_keys_file = os.path.join(dep_db_dir, 'very_militant_data.keys')
    #dependency_train_file = os.path.join(dep_db_dir, 'very_militant_data.train')

    print 'dependency train file: %s' % (dependency_train_file,)
    ## Created
    ## Various versions dependency_train_file with the same name
    ## but with suffixes and extensions added.
    lemma_dependency_train_file = os.path.join(dep_db_dir, 'lemma_dependency_DB.train')
    nearest_neighbor_file = 'nearest_neighbor.dict'
    mnn_file = 'mutual_nearest_neighbors.dict'
    #lemma_dependency_train_file = os.path.join(dep_db_dir, 'bnc_xml_lemma_dependency_DB.train')
    #nearest_neighbor_file = 'bnc_xml_nearest_neighbor.dict'
    #mnn_file = 'bnc_xml_mutual_nearest_neighbors.dict'

    ###############################################################
    ### Major activities for this module  #########################

    ###############################################################
    #print 'Getting deprel counts'
    #convert_to_mi(dependency_train_file, lemmatizer=True, cts_only=True)
    #print 'Deprel cts gotten'
    # Make codebooks from the dependency_DB.keys file.  Not yet used for anything.
    # codebooks = make_dependency_db_codebooks(dependency_keys_file)
    # Generalized better version of C{make_dependency_db_codebooks}
    from dep_rel_info import get_deprel_info, print_deprel_profile
    #print 'Getting deprel info'
    #get_deprel_info (dependency_keys_file)
    #print 'Got deprel info'
    # print_deprel_profile(('gods','NNS'),word_deprel_ct_dict)
    # print_deprel_profile(('god','n'),deprel_dict=lexeme_deprel_ct_dict)
    # Get per lexeme deprel freq info by deprel name (SBJ,OBJ, etc.)
    # print 'sorting deprel freqs'
    # (token_frequencies, mod_frequencies, best_n, best_ratio_n, best_v, best_ratio_v, \
    # avg_n_ratio, avg_v_ration, pos_ctr, mod_pos_ctr, freq_pos_ctr) = sort_deprel_freqs () 
    # print 'deprel freq sorted'
    
    # Raw deprel counts from parser converted to MI vectors.
    # By default, these are normalized to unit size.
    # Saved in file C{dependency_train_file}, with suffixes and extensions
    # to preserve name uniqueness, to serve as our base model henceforth.
    # To turn off normalization, add flag normalize=False.
    prune_neg = False
    print 'Making pruned model'
    (rel_dict, decodebook) = get_deprel_info(dependency_keys_file)
    # feat_thresh = 4,wd_threshhold=8 defaults are good used to make stanford_bnc_pruned
    # File size ratio 500415051 => 413909308
    pruned_feats = \
        make_pruned_model(dependency_train_file,dependency_keys_file,
                          decodebook,prune_neg=prune_neg,feat_threshhold=4,
                          wd_threshhold=8)
    #raise Exception
    #print 'Making dep model using unnormalized pmi'
    #convert_to_mi(dependency_train_file, normalize='',\
    #              mi_func = 'counts_to_pointwise_mutual_info',\
    #              dependency_keys_file = dependency_keys_file,\
    #              suffix = '', ext='_pmi_no_norm')
    #print 'Making dep model using hell normalized pmi'
    #convert_to_mi(dependency_train_file, normalize='hell',\
    #              mi_func = 'counts_to_pointwise_mutual_info_nonneg',\
    #              dependency_keys_file = dependency_keys_file,\
    #              suffix = '', ext='pmi_hell')
    #counts_to_normalized_pointwise_mutual_info 
    #print 'Making dep model using unnormalized pmi'
    #convert_to_mi(dependency_train_file, normalize='',\
    #              mi_func = 'counts_to_pointwise_mutual_info_nonneg',\
    #              dependency_keys_file = dependency_keys_file,\
    #              suffix = '', ext='_pmi_no_norm')
    #print 'Making dep model using normalized npmi'
    #convert_to_mi(dependency_train_file, normalize='l2',\
    #              mi_func = 'counts_to_normalized_pointwise_mutual_info',\
    #              dependency_keys_file = dependency_keys_file,\
    #              suffix = '', ext='_npmi')    
    # print 'Making dep model using cos and mi'
    # convert_to_mi(dependency_train_file, \
    #               mi_func = 'counts_to_mutual_info', 
    #               suffix = '_cos', ext='_mi')
    # print 'Making dep model using dot_prod and mi'
    # convert_to_mi(dependency_train_file,\
    #               mi_func = 'counts_to_mutual_info', normalize='',\
    #               suffix='_dot_prod',ext='_mi')
    # convert_to_mi(dependency_train_file,\
    #              mi_func = 'sqrt_freq', normalize='l1',\
    #              suffix='_dot_prod',ext='_prob_sqrt')
    # print 'Making dep model using dot_prod and pmi and lemma info'
    # convert_to_mi(dependency_train_file,\
    #               mi_func = 'counts_to_pointwise_mutual_info', normalize='',\
    #               suffix='_dot_prod',ext='_pmi',lemmatizer=True, save_cts = True)
    # print 'Making dep model using cos and undiscounted pmi'
    # convert_to_mi(dependency_train_file,\
    #             mi_func = 'counts_to_undiscounted_pointwise_mutual_info', normalize='l2',\
    #             suffix='_cos',ext='_pmi_no_disc',save_cts=True)
    #print 'Making dep model using cos and positive pmi'
    #convert_to_mi(dependency_train_file,\
    #             mi_func = 'counts_to_pointwise_mutual_info_nonneg', normalize='l2',\
    #             suffix='_cos',ext='_pmi_nonneg',save_cts=True)
    #print 'Making dep model using dotprod and conditional prob' NB: Equiv to _irad_prob !
    #convert_to_mi(dependency_train_file,\
    #             mi_func = 'counts_to_pointwise_mutual_info_nonneg', normalize='l1',\
    #             suffix='_cos',ext='_prob',save_cts=True)
    #print 'Making dep model using normalize =none and t_test rescaled'
    #convert_to_mi(dependency_train_file,\
    #             mi_func = 'counts_to_t_test_rescaled', normalize='',\
    #             suffix='',ext='_t_test_rescaled_no_norm')
    #print 'Making dep model using normalize = l2 and t_test_nonneg'
    #convert_to_mi(dependency_train_file,\
    #             mi_func = 'counts_to_t_test_nonneg', normalize='l2',\
    #             suffix='',ext='_t_test_nonneg')
    #print 'Making dep model using no normalize and t_test'
    #convert_to_mi(dependency_train_file,\
    #             mi_func = 'counts_to_t_test', normalize='',\
    #             suffix='_dot_prod',ext='_t_test')
    #print 'Making dep model using pmi'
    #convert_to_mi(dependency_train_file,\
    #            mi_func = 'counts_to_pointwise_mutual_info', normalize='',\
    #            suffix='_dot_prod',ext='_pmi_no_norm',store_sqs=True)
    #print 'Making dep model using qw poisson norm'
    #convert_to_mi(dependency_train_file,\
    #            mi_func = 'counts_to_qw_poisson', normalize='l2',\
    #            suffix='',ext='_qw_poiss',store_sqs=False)
    #print 'Making dep model using log pmi no norm [l2 norming will have no effect]'
    #convert_to_mi(dependency_train_file,\
    #            mi_func = 'counts_to_log_pmi', normalize='',\
    #            suffix='',ext='_log_pmi_no_norm',store_sqs=False)
    #print 'Making dep model using hd_pmi no_norm'
    #convert_to_mi(dependency_train_file,\
    #            mi_func = 'counts_to_hd_pointwise_mutual_info', normalize='',\
    #            suffix='',ext='_hd_pmi_no_norm',store_sqs=False)
    #print 'Making dep model using Lin PMI norm'
    #convert_to_mi(dependency_train_file,\
    #            mi_func = 'counts_to_lin_pointwise_mutual_info', normalize='l2',\
    #            suffix='',ext='_lpmi',store_sqs=False)
    #  Same with LEMMA INFO!
    # convert_to_mi(dependency_train_file,\
    #             mi_func = 'counts_to_pointwise_mutual_info', normalize='l2',\
    #             suffix='_cos',ext='_pmi',lemmatizer=True,cts_only=True)
    # get_deprel_info(os.path.join(dep_db_dir, 'dependency_DB.keys'))
    # print_deprel_profile(('god','n'),deprel_dict=lexeme_deprel_ct_dict)
    #
    # # Apr 9 Testing IRad
    # convert_to_mi(dependency_train_file,\
    #             mi_func = 'conditional_probs', normalize='l1',\
    #             suffix='_irad',ext='_prob')

    ###############################################################
    # print 'Reading model dependency file with mi numbers'
    ## When called from toplevel, the following fn is usually used to
    ## read in a pmi model, so the C{count_file} flag is False.
    #print 'Reading in a count model to make a lemma dict'
    #(word_deprel_ct_dict, word_ct_dict, deprel_ct_dict, lemma_dict, total_deprels) = \
    #                      read_dependency_file(dependency_train_file, count_file=True, \
    #                                           ext = '', prefix='lemma_')
    #print 'Reading in a count model'
    #(word_deprel_ct_dict, word_ct_dict, deprel_ct_dict, lemma_dict, total_deprels) = \
    #                     read_dependency_file(dependency_train_file, count_file=True, \
    #                                           ext = '')
    # vocab = dep_file_utils.get_vocab(os.path.join(dep_db_dir, 'dependency_DB.train'),'NN',(3000,3500))
    # print 'Creating a pruned model. Done just once!'
    #output_dependency_dict(word_deprel_ct_dict, dependency_train_file, count_file=True,\
    #                       prune = True, new_wt_file = 'dependency_DB_pruned.train',\
    #                       write_dir = dep_db_dir)
    #print 'Reading in a cosine model'
    #(word_deprel_ct_dict, word_ct_dict, deprel_ct_dict, lemma_dict, total_deprels) = \
    #                      read_dependency_file(dependency_train_file, count_file=False, \
    #                                           suffix = '', ext = '_pmi')
    # print 'Reading in a dot prod prob_sq_rt model'
    # (word_deprel_ct_dict, word_ct_dict, deprel_ct_dict, lemma_dict, total_deprels) = \
    #                      read_dependency_file(dependency_train_file, count_file=False, \
    #                                           ext = '_prob_sqrt')
    # print ''Reading in a lemma pmi model for cosine'
    #(lexeme_deprel_ct_dict, lexeme_ct_dict, deprel_ct_dict, lemma_dict, total_deprels) = \
    #                      read_dependency_file(dependency_train_file, count_file=False, \
    #                                           ext = '_pmi_no_norm',prefix='lemma_')
    ## read in a IRad appropriate prob model
    #(lexeme_deprel_ct_dict, lexeme_ct_dict, deprel_ct_dict, lemma_dict, total_deprels) = \
    #                      read_dependency_file(dependency_train_file, count_file=False, \
    #                                           ext = '_prob')

    # Loading in an mi model gives us no info about word counts. Back to counts file
    # to get that info.
    # word_ct_dict = dep_file_utils.get_word_counts(dependency_train_file)
    # import cluto_prep
    # cluto_prep.write_cluto_file(dependency_train_file,word_deprel_ct_dict)
    # cluto_prep.write_cluto_file(dependency_train_file,word_deprel_ct_dict, pos='NN',top_n=500)
                                
    ## If C{nearest_neighbor_dict} has already been computed for this model,
    ## but has been stored in a file.
    # print 'Reading Nearest neighbor file'
    # read_nn_file(nearest_neighbor_file,suffix='_cos', ext='_pmi')

    
    ##################################################################################################
    ################################to##################################################################
    ###  find_nearest_neighbors ######################################################################
    ##################################################################################################
    ##################################################################################################
    # When C{nearest_neighbor_dict} does not already exist for this model.
    # caution: specifying a "top_n" presupposes accurate word cts are available.
    # Find nearest neighbors with same parts of speech for 500 most freq words.
    #find_nearest_neighbors (dependency_train_file, nearest_neighbor_file=nearest_neighbor_file, \
    #                        top_n=500, preserve_pos=True, sim_func='baseline', \
    #                        pos='NN', ext='_poisson_norm', verbose=True)
    # Find nearest neighbors with same parts of speech for 500 most freq words.
    #find_nearest_neighbors (dependency_train_file, nearest_neighbor_file=nearest_neighbor_file, \
    #                        top_n=500, preserve_pos=True, sim_func='dot_prod', \
    #                        pos='NN', ext='_poisson_norm', verbose=True)
    #find_nearest_neighbors (dependency_train_file, nearest_neighbor_file=nearest_neighbor_file, \
    #                        top_n=500, preserve_pos=True, sim_func='dot_prod', \
    #                        pos='NN', ext='_pmi_norm', verbose=True)
    #dot_prod_t_test_no_norm = reset_rel_nearest_neighbor('NN')
    #find_nearest_neighbors (dependency_train_file, nearest_neighbor_file=nearest_neighbor_file, \
    #                        top_n=(500,1000), preserve_pos=True, sim_func='lin_no_neg', \
    #                          pos='NN', ext='_t_test_no_norm', verbose=True)
    #lin_no_neg_t_test_no_norm = reset_rel_nearest_neighbor('NN')
    #find_nearest_neighbors (dependency_train_file, nearest_neighbor_file=nearest_neighbor_file, \
    #                        top_n=(500,1000), preserve_pos=True, sim_func='dice_dagger', \
    #                          pos='NN', ext='_t_test_no_norm', verbose=True)
    # Hellinger distance.  dot prod combined with prob sqrts
    # find_nearest_neighbors (dependency_train_file, nearest_neighbor_file=nearest_neighbor_file, \
    #                        top_n=500, preserve_pos=True, sim_func='dot_prod', \
    #                        pos='NN', ext='_prob_sqrt', verbose=True)
    # Lin sim. counts_to_info weights.
    # print 'Finding lin sim nearest neighbors for 500 most freq nouns'
    # find_nearest_neighbors (dependency_train_file, nearest_neighbor_file=nearest_neighbor_file, \
    #                         top_n=500, preserve_pos=True, sim_func='lin_no_neg', \
    #                        pos='NN', ext='_pmi_no_norm', verbose=True)
    # print 'Finding jaccard no_norm nearest neighbors for 500 most freq nouns'
    #find_nearest_neighbors (dependency_train_file, nearest_neighbor_file=nearest_neighbor_file, \
    #                        top_n=500, preserve_pos=True, sim_func='dice_dagger', \
    #                        pos='NN', ext='_pmi_no_norm', verbose=True)
    #find_nearest_neighbors (dependency_train_file, nearest_neighbor_file=nearest_neighbor_file, \
    #                        top_n=500, preserve_pos=True, sim_func='dice_dagger_no_neg', \
    #                        pos='NN', ext='_pmi_nonneg', verbose=True)
    # print 'Finding cos nearest neighbors for 500 most freq nouns'
    # find_nearest_neighbors (dependency_train_file, nearest_neighbor_file=nearest_neighbor_file, \
    #                        top_n=500, preserve_pos=True, pos='NN', ext='_pmi')
    # print 'Finding cos t_test nearest neighbors for 500 most freq nouns'
    # We must have created or read in a model using mi_func = 'counts_to_t_test'
    # find_nearest_neighbors (dependency_train_file, nearest_neighbor_file=nearest_neighbor_file, \
    #                        top_n=500, preserve_pos=True, pos='NN', suffix='_cos', ext='_t_test')
    #print 'Finding dot_prod t_test nearest neighbors for 500 most freq nouns'
    # We must have created or read in a model using mi_func = 'counts_to_t_test', with no normalization
    # find_nearest_neighbors (dependency_train_file, nearest_neighbor_file=nearest_neighbor_file, \
    #                        top_n=500, preserve_pos=True, pos='NN', ext='_t_test_rescaled_no_norm',verbose=True)
    # print 'Finding jaccard pmi nearest neighbors for 500 most freq nouns'
    # find_nearest_neighbors (dependency_train_file, nearest_neighbor_file=nearest_neighbor_file, \
    #                        top_n=500, preserve_pos=True, pos='NN', suffix='_jaccard', ext='_pmi',\
    #                        sim_func='jaccard', verbose=True)
    # print 'Finding jaccard t_test (no norm) nearest neighbors for 500 most freq nouns'
    # We must have created or read in a model using mi_func = 't_test', with no normalization
    # find_nearest_neighbors (dependency_train_file, nearest_neighbor_file=nearest_neighbor_file, \
    #                         top_n=500, preserve_pos=True, pos='NN', suffix='_jaccard', ext='_t_test_no_norm',\
    #                        sim_func='jaccard', verbose=True)
    #    # Find nearest neighbors for 500 most freq VBD
    # find_nearest_neighbors (dependency_train_file, nearest_neighbor_file=nearest_neighbor_file, \
    #                        top_n=500, preserve_pos=True, pos='VBD', suffix='_cos', ext='_pmi')
    # Find nearest neighbors for 500 most freq 'v' lexemes;
    # assumes lexeme_ct_dict and lexeme_deprel_ct_dict
    # contain pmi numbers for lexemes, which in turn assumes they were created by passing
    # C{convert_to_mi} the lemmatizer=True flag.
    # find_nearest_lexeme_neighbors (dependency_train_file, lexeme_ct_dict, lexeme_deprel_ct_dict, \
    #                               nearest_neighbor_file, top_n=500, pos='v',\
    #                               suffix='_cos', ext='_pmi', prefix='lemma_')

    # Apr 7
    #find_nearest_neighbors(lemma_dependency_train_file, nearest_neighbor_file,top_n=500, \
    #                        suffix='_cos',ext='_pmi',preserve_pos=True, pos='n',\
    #                      deprel_dict=lexeme_deprel_ct_dict)

    # Apr 9 Testing IRad
    ## after having done
    ## dump_rel_nearest_neighbor_dict ('partial'+nearest_neighbor_file, pos='NN',\
    ##                                   suffix = '_IRad', ext = '_prob')
    # nn_dict = {}
    # rel_nearest_neighbor_dict['NN'] = nn_dict
    # read_nn_file('rel__partialnearest_neighbor_NN_IRad_prob.dict',nn_dict=nn_dict)
    # find_nearest_neighbors(dependency_train_file, nearest_neighbor_file,top_n=500, \
    #                       suffix='_IRad',ext='_prob',preserve_pos=True, pos='NN',\
    #                       sim_func='IRad',verbose=True)

    
    ###############################################################
    ### End Major activities for this module  #####################
    ###############################################################

    #### Fun things to try  ######################
    
    # nn1 = find_nearest_neighbor(('church','NN'))
    # find_nearest_n(('wants','VBZ'), 10)
    # make_neighborhood_graph(('be','VB'),5,dot_file=True)
    # make_neighborhood_graph(('wants','VBZ'),5,preserve_pos=True,dot_file=True)

    #rank_dict = {}
    #print len(rank_dict)
    #ct_dict = make_ct_dict_fill_rank_dict(dependency_train_file, rank_dict)
    #template = 'rel_nearest_neighbor_NN_3000_3500_%s.dict'
    #dict_dir = os.path.join('july_results','alt_3000')
    #dict_dir = 'bulba_nn_files'
    #dict_dir = os.path.join('july_results','dot_prod_sq_rt')
    #dict_dir = os.path.join('july_results','dot_min')
    #dict_dir = os.path.join('july_results','dice_dagger')
    #dict_dir = os.path.join('july_results','dot_prod_sq_rt')
    #dict_dir = os.path.join('july_results','log_scp')
    # $\sqrt{\text{dot prod}}$ & PMI & 0.515\\
    # dot min & PMI   &  0.509\\
    # dot prod & PMI no norm & 0.494 \\
    # dot min & PMI no norm  &  0.480\\
    # $\sqrt{\text{dot prod}}$ & t-test2 & 0.475\\
    # $\sqrt{\text{dot prod}}$ & log scp & 0.463\\
    # $\sqrt{\text{dot prod}}$ &  pmi no norm & 0.456\\
    # $\sqrt{\text{dot prod}}$ &  dice wt & 0.451\\
    # dice dot prod & PMI no norm & 0.441 \\
    # \dice & PMI no norm &  0.440  \\
    #(rank_dict_dict, rank_dict, rank_results) = \
    #          get_rank_results (dict_dir, rank_dict,dependency_train_file)
    #sim_systems = rank_results.keys()
    #attractor_lists = make_attractor_lists(sim_systems,rank_dict_dict)
    #for sim_s in sim_systems:
    #    print
    #    print_attractors(sim_s, attractor_lists)
    #nn_file = os.path.join('test_nn_file', \
    #                       'optimal_nearest_nneighbors_3000_3500.dict')
    #nn_dict = {}
    #read_nn_file(nn_file,nn_dict=nn_dict)
    
        
        
        
