"""
This file hooks wordnet up to BNC for evaluation
of word similarity measures using The semantic similarity
metrics implemented in Wordnet.  All use the Wordnet hyponym
hierarchy, but some, the ones of interest here, build
in an information content measure for the concepts.
This is information theory-based, hence probabilistic,
and therefore a corpus is needed.  
because we have used a set of word similarity functions
to collect "nearest neighbors" using BNC,
so we will also use BNC to provide the information
content statistics for WN.

Much of this has been implemented in NLTK.
To find more about nltk interface to Wordnet

 http://nltk.googlecode.com/svn/trunk/doc/howto/wordnet.html

To find out more about creating an nltk corpus (which is easy):
http://nltk.googlecode.com/svn/trunk/doc/howto/corpus.html#writing-new-corpus-readers.

Here we create an NLTK corpus for BNC, then link it to
Wordnet, then evaluate nearest neighbor pairs of words, where
nearest has been determined by applying some
distribution-based similarity function to BNC.

    >>> root = os.path.join(corpus_dir, 'one_article_files')
    >>> bnc = TaggedCorpusReader(root, '.*', sep='_')
    >>> wds = bnc.words()
    >>> print 'The BNC Corpus is %d words long.' % (len(wds),)
    >>> bnc_ic = wn.ic(bnc, False, 0.0)

Note one can use Pedersen's supplied bnc ic instead.  Results appear
indistinguishable. Disappointingly doing this does NOT appear
to be faster.  Apparently Python takes a long to load the '.dat' file.

    >>> from nltk.corpus import wordnet_ic
    >>> bnc_ic = wordnet_ic.ic('ic-bnc-add1.dat',False,0.0)
    
For our purposes here the only thing to test for the BNC corpus is::

  >>> brown.words()
  ['The', 'Fulton', 'County', 'Grand', 'Jury', 'said', ...]

Other methods which also work::

  >>> brown.root 
  FileSystemPathPointer('.../corpora/brown')
  >>> brown.sents() 
  [['The', 'Fulton', 'County', 'Grand', ...], ...]
  >>> brown.paras() 
  [[['The', 'Fulton', 'County', ...]], [['The', 'jury', ...]], ...]
  >>> brown.tagged_words() 
  [('The', 'AT'), ('Fulton', 'NP-TL'), ...]
  >>> brown.tagged_sents() 
  [[('The', 'AT'), ('Fulton', 'NP-TL'), ('County', 'NN-TL'), ...], ...]
  >>> brown.tagged_paras() 
  [[[('The', 'AT'), ...]], [[('The', 'AT'), ...]], ...]

Note: need the words() method to create an information file resource
for using similarity measures in WordNet.

"""

## Path to the nltk tagged corpus reader
## /Library/Python/2.6/site-packages/nltk/corpus/reader/tagged.py
import sys, os.path, time
import re
from nltk.corpus import TaggedCorpusReader
#nltk.corpus.reader.wordnet.WordNetError
from nltk.corpus import wordnet as wn
# May be needed by verbose_report, because of call to a FUNCTION
# (not a method) in nltk.corpus.reader.wordnet, namely _lcs_ic.
import nltk.corpus.reader.wordnet
import dependency_db, dep_file_utils
from bnc_wn_interface import bnc_word_to_wn_word, bnc_pos_to_wn_pos

def eval_experiment(ic, wn_simfn, pos='', db_simfn='', space='', nn_dict_file='', \
                    nn_dict_dir='', write_dir='', verbose=True, prefix='rel_pp_nearest_neighbor'):
    """
    Read a nearest neighbor file C{nn_dict_file} from the dependency db directory,
    C{nn_dict_dir}. Score those nearest neighbors relatve to  wordnet similarity measure
    C{simfn}, using information corpus, C{ic}.
    """
    global top_wn_subsumer
    #  get the most general noun synset
    # Used in wn_sim_score, the pairwise synset,synset similarity function
    # Unknown words are those whose most specific subsumer is top_wn_subsumer
    top_wn_subsumer = wn.all_synsets('n').next()


    if not nn_dict_file:
        nn_dict_file = dependency_db.make_nn_file_name(pos, db_simfn, space, prefix=prefix)
    nn_dict = read_pair_scores (os.path.join(nn_dict_dir,nn_dict_file))
    (base, ext) = os.path.splitext(nn_dict_file)
    if nn_dict:
        score_nn_dict(nn_dict, ic, base + wn_simfn + ext,\
                      simfn=wn_simfn, write_dir=write_dir, verbose=verbose,\
                      n_dict_file=nn_dict_file)

def eval_directory(ic, data_dir, write_dir='', wn_simfn='lin_similarity',log_file=True, verbose=True):
    global top_wn_subsumer
    top_wn_subsumer = wn.all_synsets('n').next()
    
    file_list = os.listdir(data_dir)
    for fn in file_list:
        (base,ext) = os.path.splitext(fn)
        if ext == '.dict':
            nn_dict = read_pair_scores (os.path.join(data_dir,fn))
            if nn_dict:
                try:
                    score_nn_dict(nn_dict, ic, report_file= base + wn_simfn + ext, simfn = wn_simfn, \
                                  write_dir=write_dir, verbose = verbose, nn_dict_file=fn, log_file=log_file)
                except ZeroDivisionError:
                    print '%s: zero division error!' % (fn,)
                    continue

## Warning!!! Platform specific!!!
min_float = 9.88131291682e-323
def score_nn_dict (nn_dict, ic, report_file='', simfn='lin_similarity', write_dir='', verbose=False,\
                   nn_dict_file='',log_file=True):
    if report_file:
        ofh = dep_file_utils.open_and_make_backup(os.path.join(write_dir,report_file))
        # ofh = open(os.path.join(write_dir,report_file),'w')
    if log_file:
        log_fh = open(os.path.join(write_dir,'wordnet_evals.log'),'a')
        print >> log_fh, '========================================================================='
        print >> log_fh, nn_dict_file, time.asctime()
        print >> log_fh, '========================================================================='
    new_nn_dict = {}
    summed_scores = 0
    score_list = []
    ct = 0
    neg_ctr = 0
    zero_ctr = 0
    midline_ctr = 0
    grade_a_ctr = 0
    if simfn == 'lin_similarity':
        grade_a = .9
        midline = .5
    elif simfn == 'jcn_similarity':
        grade_a = .67
        midline = .10
        #grade_a = .9
        #midline =.5
        
    for (lex1,lex2) in nn_dict:
        ct += 1
        score = wn_sim_score(lex1,lex2, ic, simfn=simfn, verbose = verbose)
        if simfn == 'jcn_similarity':
            # switch to bounded version
            #score = score/(1 + score)
            print '%.3e' % score
            pass
        score_list.append(score)
        if score > 0:
            ## Negs mean unknown words, treated as 0 sims.
            summed_scores += score
        if score < midline:
            midline_ctr += 1
            if score < 0:
                neg_ctr += 1
                #  JMG Why was this here?
                #  wn_sim_score(lex1,lex2, ic, simfn=simfn, verbose = verbose)
            if score < min_float:
                zero_ctr += 1
        elif score > grade_a:
            grade_a_ctr += 1
        new_nn_dict[lex1] = (lex2, score)
        if report_file:
            print >> ofh, lex1, (lex2, score)
    nonneg_ct = ct - neg_ctr
    ## EDIT THIS to affect conrath outputs!!!!
    #default_fns = ['lin_similarity','jcn_similarity']
    default_fns = ['lin_similarity']
    if simfn == 'jcn_similarity':
        score_str = 'Average sim score: %.3e Average known word sim score %.3e'
    else:
        score_str = 'Average sim score: %.3f Average known word sim score %.3f'
    if simfn in default_fns:
        report_str = score_str % (summed_scores/float(ct), summed_scores/float(nonneg_ct))
    elif simfn == 'jcn_similarity':
        (median,nonneg_median) = find_score_medians(score_list,ct)
        report_str = 'Average sim score: %.3f Average known word sim score %.3f' % (median,nonneg_median)
        # report_str = 'Average sim score: %.3f Average known word sim score %.3f' % (summed_scores/float(ct), summed_scores/float(nonneg_ct))
    else:
        raise Exception, 'Unknown simfn: %s' % (simfn,)
    report_str2 = '>%.2f: %d  >%.2f: %d  =0: %d' % (grade_a, grade_a_ctr, midline, ct - midline_ctr, zero_ctr,)                                                 
    if report_file:
        print_digest_info (nn_dict_file, report_str, report_str2, handle=ofh)
        ofh.close()
    if log_file:
        print_digest_info (None, report_str, report_str2, handle=log_fh)
        log_fh.close()
    # Print the digest info to stdout
    print_digest_info (nn_dict_file, report_str, report_str2)

import re
## Match ((word,pos), ((neighbor,pos), score))
nn_dd_line = re.compile(r'(\(.+\))\s*\((\(.+\))\s*,\s*(.+)\)')

def read_pair_scores (nn_file):
    fh = open(nn_file,'r')
    match_found = False
    score_dict = {}
    for line in fh:
        line = line.strip()
        match = nn_dd_line.match(line)
        if match:
            match_found = True
            groups = match.groups()
            (lex1,lex2,score) = (eval(groups[0]),eval(groups[1]),eval(groups[2]))
            score_dict[(lex1,lex2)] = score
    if not match_found:
        print 'Not a nearest neighbor file!'
    fh.close()
    return score_dict

def print_digest_info (nn_dict_file, report_str, report_str2, handle=sys.stdout):
    if nn_dict_file:
        print >> handle, 'Nearest neighbor file %s' % (nn_dict_file,)
    print >> handle, report_str
    print >> handle, report_str2
    
def wn_sim_score(lex1,lex2, ic, simfn='lin_similarity', verbose = False):
    """
    Default val for C{simfn} is lin_similarity. Note that only Wordnet sim fns that
    use an information corpus argument (here, C{ic}) will work.  Currently these
    are lin_similarity, jcn_similarity, and res_similarity.

    Unknown words are those whose most specific subsumer is top_wn_subsumer, a global
    Comments on using lin sim score on Wordnet.  May apply to other ic measures equally well.

    >>> wn_sim_score(('advice','NN'),('guidance','NN'), bnc_ic, simfn = 'lin_similarity', verbose = True)

     advice.n.01:  a proposal for an appropriate course of action
     Hypernyms: [ proposal.n.01,  message.n.02,  communication.n.02,  abstraction.n.06,  entity.n.01,  ]
     
     guidance.n.01:  something that provides direction or advice as to a decision or course of action
     Hypernyms: [ message.n.02,  communication.n.02,  abstraction.n.06,  entity.n.01,  ]
     
     Least common hypernyms:  message.n.02
     advice.n.01 ic: 7.5687 guidance.n.01 ic: 8.7013 message.n.02 ic: 3.9498

    The problem is that what dominates "advice" and "guidance" should be something far more specific
    than "message", which is a VERY general node with an accordingly low info value, so the similarity
    of "advice" and "guidance" gets way undervalued. So being at the mercy of the hierarchy structure
    is a problem.  Especially when combined with the probabilities.  But of course the Wordnet hierarchy
    structure is itself at the mercy of the English lexicon.  There just isnt an obvious WORD
    that immediately dominates advice and guidance.  Framenet beckons.

    Other issues: Some abstract nodes get surprisingly high relative scoores.
    ('rise', 'NN')   ('reduction', 'NN') => .77740  because the lch change_of_magnitude.n.01 has ic: 7.0943
    
    while

    ('rise', 'NN') ('number', 'NN') => .7544   because lch  amount.n.02 has ic: 6.9112

    """
    global top_wn_subsumer
    
    unknown_wn_word_str = 'WN: Unknown word: %s %s'

    ((word1,bnc_pos1), (word2,bnc_pos2)) = (lex1, lex2)
    (wn_pos1,wn_pos2) = (bnc_pos_to_wn_pos(bnc_pos1), bnc_pos_to_wn_pos(bnc_pos2))
    (word1, word2) = (bnc_word_to_wn_word(word1), bnc_word_to_wn_word(word2))
    max_sim = 0
    (s1_max,s2_max) = (top_wn_subsumer, top_wn_subsumer)
    (ss_list1,ss_list2) = (wn.synsets(word1,pos=wn_pos1), wn.synsets(word2,pos=wn_pos2))
    unknown_test_word = False
    if ss_list1 and ss_list2:
        for s1 in ss_list1:
                s1_sim = getattr(s1,simfn)
		for s2 in ss_list2:
                    try:
                        score = s1_sim(s2,ic)
                    except nltk.corpus.reader.wordnet.WordNetError:
                        return max_sim                        
                    if score >= max_sim:
                        max_sim = score
                        s1_max = s1
                        s2_max = s2
    elif ss_list1:
        max_sim = -1
        if verbose:
            print unknown_wn_word_str % (word2,wn_pos2)
    else:
        max_sim = -1
        if verbose:
            unknown_test_word = True
            print unknown_wn_word_str % (word1,wn_pos1)
            if not ss_list2:
                print unknown_wn_word_str % (word2,wn_pos2)
    if verbose:
	    verbose_report(s1_max,s2_max,max_sim,lex1, lex2, ic, unknown_test_word)
    return max_sim

def verbose_report(s1,s2,score,lex1,lex2,ic,unknown_test_word):
	hyp = lambda s:s.hypernyms()
	print '************************************************'
	print lex1, lex2
	print
        if unknown_test_word:
            print 'Unknown test word!'
	for ss in [s1,s2]:
		print '%s: ' % (ss.name,),
		print ss.definition
		print 'Hypernyms: [',
		for sk in ss.closure(hyp):
		    print sk.name+', ',
		print ']'
		print
	print 'Least common hypernyms: ',
	lcs_list = s1.lowest_common_hypernyms(s2)
	for ss in lcs_list:
		print ss.name
        try:
            (s1_ic, s2_ic, lcs_ic) = nltk.corpus.reader.wordnet._lcs_ic(s1, s2, ic)
        except nltk.corpus.reader.wordnet.WordNetError:
            return
        # Doesnt work, because of some nltk trickery with names.
	# (s1_ic, s2_ic, lcs_ic) = wn._lcs_ic(s1, s2, ic)
	print '%s ic: %.4f' % (s1.name,s1_ic,), '%s ic: %.4f' % (s2.name,s2_ic,),
        if lcs_list:
            '%s ic: %.4f' % (lcs_list[0].name,lcs_ic,)
        else:
            print 'No lowest common hypernyms!'
	print 'Sim score: %.4f' % (score,)
	print


def find_score_medians(score_list,ct):
    score_list.sort()
    nonneg_score_list = [s for s in score_list if s >= 0]
    nonneg_score_list.sort()
    nonneg_ct = len(nonneg_score_list)
    median = score_list[ct/2]
    nonneg_median = nonneg_score_list[nonneg_ct/2]
    return (median,nonneg_median)

def alt_find_score_medians(score_list,ct):
    sl = []
    nonneg_score_list = [] 
    for s in score_list:
        if s >=  0:
            new_s = float(s)/(1 +s)
            sl.append(new_s)
            nonneg_score_list.append(new_s)
        else:
            sl.append(s)
    sl.sort()
    nonneg_score_list.sort()
    nonneg_ct = len(nonneg_score_list)
    median = sl[(ct/2)]
    nonneg_median = nonneg_score_list[(nonneg_ct/2)]
    return (median,nonneg_median)

def wn_find_nearest_neighbors (vocab, ic, simfn = 'lin_similarity', \
                               test_set=None, test_file=None,nn_dict_dir='',\
                               verbose=False):
    global top_wn_subsumer, wn_nn_dict
    top_wn_subsumer = wn.all_synsets('n').next()
    assert test_set or test_file
    if not test_set:
        test_list = [k[0] for k in \
                      read_pair_scores (os.path.join(nn_dict_dir,\
                                                     test_file)).iterkeys()]
        test_set = set(test_list)
    wn_nn_dict = {}
    test_ctr = 0
    for lex in test_set:
        test_ctr += 1
        best = -1
        best_lex = ''
        for lex2 in vocab:
            if lex2 == lex:
                continue
            score = wn_sim_score(lex,lex2, ic, simfn = simfn, verbose = False)
            if score > best:
                best = score
                best_lex = lex2
        wn_nn_dict[lex] = (best_lex, best)
        if verbose:
            if test_ctr%100 == 0:
                print '*** %d done! ***' % (test_ctr,)
            print lex, best_lex, best
    total = 0.0
    ctr = 0
    for (lexn, v) in wn_nn_dict.itervalues():
        ctr += 1
        total += v
    print 'Avg: %.3f' % (total/ctr,)
    return wn_nn_dict

    

if __name__ == '__main__':
    import sys
    
    ############################################################################
    ############################################################################
    ### BNC
    ############################################################################
    ############################################################################

    ############################################################################
    ## Set this var!
    ############################################################################
    load_bnc = True
    home = os.environ['HOME']
    root_dir = os.path.join(home,'ext')
    corpus_dir = os.path.join(root_dir, 'corpus','bnc')
    # root = os.path.join(home,'ext','corpus','bnc','one_article_files','A')
    root = os.path.join(corpus_dir, 'one_article_files')
    if load_bnc:
        bnc = TaggedCorpusReader(root, '.*', sep='_')
    # wds = bnc.words()
    # print 'The BNC Corpus is %d words long.' % (len(wds),)

    ############################################################################
    ############################################################################
    ### Hooking Wordnet up to BNC
    ############################################################################
    ############################################################################

    if load_bnc:
        print 'Loading bnc'
        bnc_ic = wn.ic(bnc, False, 0.0)
        print 'bnc loaded'
    #  Note using Pedersen's provided ic BNC file instead. NB: Entire BNC.
    #  seems to make NO difference.  Numbers the same.  Same speed!
    #from nltk.corpus import wordnet_ic
    #bnc_ic = wordnet_ic.ic('ic-bnc-add1.dat',False,0.0)
    
    # S = wn.synset
    # print S('dog.n.01').lin_similarity(S('cat.n.01'), bnc_ic)
    
    ############################################################################
    ############################################################################
    ### Dependency DB  nearest neighbors + Evaluation wrt Wordnet
    ############################################################################
    ############################################################################

    # Being set in PYTHONSTARTUP (see .bash_profile file for location)
    # dep_db_dir = os.path.join(home,'Desktop','SWIDSAS','dependency_DB')
    
    # sys.path.append(dep_db_dir)

    write_dir = os.path.join('lin_sim_score_results','all')

    ##  nearest neighbor dict with nearest neighbor of x calculated within the set of
    ##   words with same pos as x (preserve part of speech)

    ## parameters that almost never change
    wn_sim_fn = 'lin_similarity'
    dict_ext = '.dict'
    prefix = 'rel_pp_nearest_neighbor'
    verbose = True

    ## parameters that change occasionally
    pos = 'NN'

    ## The two parameters that change in most experiments: db_simfn and space
    #db_simfn = 'IRad'
    #db_simfn = 'cos_no_neg'
    #db_simfn = 'cos'
    #


    #db_simfn = 'cos_'
    #db_simfn = 'dot_prod'
    #db_simfn = 'lin_no_neg'
    #db_simfn = 'dice_dagger'
    #db_simfn = 'jaccard'
    #space = 'pmi_no_norm'
    #space = 't_test_no_norm'
    #space = 'pmi_nonneg'
    #space = 'pmi_no_norm'
    #space = 'prob_sqrt'
    #space = 'pmi_no_disc'

    db_simfn = '500_1000_dot_prod'
    space = 'pmi'

    # eval_experiment(bnc_ic, wn_simfn=wn_sim_fn, pos=pos, db_simfn=db_simfn, \
    #                space=space, nn_dict_dir=dep_db_dir, write_dir=write_dir, verbose=verbose, prefix=prefix)

    home = os.environ['HOME']
    working_dir = os.path.join(home,'Desktop','SWIDSAS','dependency_DB')
    os.chdir(working_dir)
    write_dir = 'test_report_file'
    nn_dict_dir = os.path.join('july_results','3000')
    test_nn_file = 'rel_pp_nearest_neighbor_NN_3000_3500_dice_dagger_pmi_no_norm.dict'
    import dep_file_utils
    print 'Getting vocab'
    vocab = dep_file_utils.get_vocab('../dep_data/dependency_DB.train','NN',20000)
    print 'vocab gotten'
    ## reverse vocab so that when looking for nearest neighbors
    ## ties are resolved in favor of the most frequent word.
    vocab = [lex for (lex,ct) in vocab[::-1]]
    wn_nn_dict = \
     wn_find_nearest_neighbors (vocab, bnc_ic, simfn = 'lin_similarity', \
                               test_set=None, test_file=test_nn_file,\
                               nn_dict_dir=nn_dict_dir,verbose=True)
    # eval_directory(bnc_ic, 'test_nn_file', write_dir = write_dir,  wn_simfn='jcn_similarity',verbose=False)
