#!/usr/bin/env python
"""
Two environment variables must be set for this evaluation code to work::

   -- DEPEND_DB points to directory in which neighbor data files generated by
   nearest neighbor experiements are kept.
   -- DEPEND_DB_CORPUS_DIR points to root of file tree for a tagged corpus.  Can
   contain subdirectories.  Used for statistics when combining information theoretic sim measures
   such as Lin's with the WordNet hierarchy.
   
"""
from nltk_bnc_corpus_reader import TaggedCorpusReader, wn, eval_experiment, eval_directory
import dependency_db

def main (corpus_file_root, nn_dict_dir='', write_dir='', weight_func='', sim_func='',\
          wn_sim_fn='lin_similarity', dict_ext='.dict',\
          prefix='rel_pp_nearest_neighbor',verbose=True,pos='NN'):
    """
    For help on options, do::

      template_call_wordnet_eval.py --help

    The two similarity functions implemented are C{lin_similarity} and
    C{jcn_similarity}.  The former is the default.
    
    For example, to use jiang_conrath on a directory C{nn_dir} and output the
    results to sister directory C{nn_dir_jcn_output}, connect to directory
    C{nn_dir} and do::
    
      -- template_call_wordnet_eval.py -e jcn_similarity -o ../nn_dir_jcn_output -d .
    """

    import os.path, sys, optparse

    ## Main options: the two parameters that change most in experiments: weight_func and sim_func
    ## Since the data has already been generated when this code is called, these options
    ## are just used to compute the name of the data file being looked up.

    option_list = [
         optparse.make_option('-w', '--weight_func', default='none',\
                              metavar='WEIGHT_FUNC', help='Function used to weight counts in word vectors'),
         optparse.make_option('-s', '--sim_func', default='none',\
                              metavar='SIM_FUNC', help='Function used to measure similarity of two word vectors'),
         optparse.make_option('-v', '--verbose', default=True, action='store_true',\
                              metavar='VERBOSITY', help='Verbose: Print sim scores for all words (True)'),
         optparse.make_option('-o', '--output_dir', default='none',\
                              metavar='DIR', help='Directory in which to place evaluation results'),
         optparse.make_option('-p', '--pos', default='NN',\
                              metavar='POS', help='Part of speech of words in evaluation word set (NN)'),
         optparse.make_option('-e', '--eval_func', default='lin_similarity',\
                              metavar='EVAL_FUNC', help='Function measuring closeness in WordNet of proposed nearest neighbors'),
         optparse.make_option('-d', '--nn_dir', default='none',\
                              metavar='DIR', help='directory of nearest neighbor files'),
        ]

    p = optparse.OptionParser(option_list=option_list)
    options, arguments = p.parse_args()

    if len(sys.argv) == 1 and sys.argv[0] == '':
        # No commandline options
        pass
    else:
        # These values overrule all others
        if options.output_dir == 'none':
            pass
        else:
            write_dir = options.output_dir
        if options.nn_dir == 'none':
            nn_dir = None
        else:
            nn_dir = options.nn_dir
        wn_simfn = options.eval_func
        verbose = options.verbose
        assert type(verbose) == bool
        pos = options.pos
        if options.weight_func == 'none':
            weight_func = None
        else:
            weight_func = options.weight_func
        if options.sim_func == 'none':
            sim_func = None
        else:
            sim_func = options.sim_func
        if nn_dir:
            if sim_func or weight_func:
                print 'Warning: nn_dir %s given. Ignoring values for sim_func (%s) or weight_func (%s).' % (nn_dir, sim_func, weight_func)
                print 'Processing all nn_files in %s' % (nn_dir,)
        
    
    ############################################################################
    ############################################################################
    ### BNC
    ############################################################################
    ############################################################################

    ## Assuming tag sep char is '_', as in BNC.  Change to '/', e.g., for Brown.
    ## Assuming filename extension for corpus files is any.  Change, e.g., to ".txt'
    
    corpus = TaggedCorpusReader(corpus_file_root, '.*', sep='_')

    ############################################################################
    ############################################################################
    ### Hooking Wordnet up to BNC
    ############################################################################
    ############################################################################

    dep_db_dir = os.environ['DEPEND_DB']

    # Check for nn file specs
    if nn_dir and os.path.isdir(nn_dir):
        # Check for existence of nearest neighbor directory early on.
        # No need to spend time loading BNC if there's no directory to evaluate.
        pass
    elif nn_dir:
        print 'No such directory: %s' % (nn_dir,)
        sys.exit()
    else:
        # Check for existence of nearest neighbor file early on.
        # No need to spend time loading BNC if there's no file to evaluate.
        nn_file = os.path.join(dep_db_dir,dependency_db.make_nn_file_name(pos, sim_func, weight_func, prefix=prefix))
        if os.path.isfile(nn_file):
            pass
        else:
            print 'No such file: %s' % (nn_file,)
            sys.exit()

    print 'Loading CORPUS (may take a few minutes)'
    corpus_ic = wn.ic(corpus, False, 0.0)
    print 'CORPUS loaded.'
    
    ############################################################################
    ############################################################################
    ### Eval experiment or directory
    ############################################################################
    ############################################################################

    if nn_dir:
        eval_directory(corpus_ic,nn_dir, write_dir=write_dir, wn_simfn=wn_simfn)
    else:
        eval_experiment(corpus_ic, wn_simfn=wn_sim_fn, pos=pos, db_simfn=sim_func, \
                        space=weight_func, nn_dict_dir=nn_dict_dir, write_dir=write_dir, \
                        verbose=verbose, prefix=prefix)

if __name__ == '__main__':
    import sys, os.path

    ############################################################################
    ############################################################################
    ### BNC
    ############################################################################
    ############################################################################

    # home = os.environ['HOME']
    # root_dir = os.path.join(home,'ext','corpus')
    # corpus_dir = os.path.join(root_dir, 'bnc')

    depend_corpus_dir = os.environ['DEPEND_CORPUS_DIR']

    #corpus_file_root = os.path.join(depend_corpus_dir, 'one_article_files')

    ###########################################################################
    ### Dependency DB  nearest neighbors + Evaluation wrt Wordnet
    ###########################################################################
    ###########################################################################


    #root = os.path.join(corpus_dir, 'one_article_files')
    #os.path.join(home, 'Desktop', 'SWIDSAS','dependency_DB')
    dep_db_dir = os.environ['DEPEND_DB']
    
    write_dir = os.path.join(dep_db_dir, 'lin_sim_score_results','all')
    #write_dir = os.path.join(dep_db_dir, 'july_results','new_jumbo')

    main(depend_corpus_dir, nn_dict_dir=dep_db_dir, write_dir=write_dir)
